#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015-2016 Douglas S. Blank <doug.blank@gmail.com>
# Copyright (C) 2016-2017 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import time
import pickle
import logging

# ------------------------------------------------------------------------
#
# Gramps Modules
#
# ------------------------------------------------------------------------
from gramps.gen.db.dbconst import (
    DBLOGNAME,
    DBBACKEND,
    KEY_TO_NAME_MAP,
    KEY_TO_CLASS_MAP,
    TXNADD,
    TXNUPD,
    TXNDEL,
    PERSON_KEY,
    FAMILY_KEY,
    SOURCE_KEY,
    EVENT_KEY,
    MEDIA_KEY,
    PLACE_KEY,
    NOTE_KEY,
    TAG_KEY,
    CITATION_KEY,
    REPOSITORY_KEY,
    REFERENCE_KEY,
)
from gramps.gen.db.generic import DbGeneric
from gramps.gen.updatecallback import UpdateCallback
from gramps.gen.lib import (
    Tag,
    Media,
    Person,
    Family,
    Source,
    Citation,
    Event,
    Place,
    Repository,
    Note,
)
from gramps.gen.lib.genderstats import GenderStats
from gramps.gen.const import GRAMPS_LOCALE as glocale

LOG = logging.getLogger(".dbapi")
_LOG = logging.getLogger(DBLOGNAME)


class DBAPI(DbGeneric):
    """
    Database backends class for DB-API 2.0 databases
    """

    def _initialize(self, directory, username, password):
        raise NotImplementedError

    def _schema_exists(self):
        """
        Check to see if the schema exists.

        We use the existence of the person table as a proxy for the database
        being new.
        """
        return self.dbapi.table_exists("person")

    def _create_schema(self):
        """
        Create and update schema.
        """
        self.dbapi.begin()

        # make sure schema is up to date:
        self.dbapi.execute(
            "CREATE TABLE person "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "given_name TEXT, "
            "surname TEXT, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE family "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE source "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE citation "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE event "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE media "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE place "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "enclosed_by VARCHAR(50), "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE repository "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE note "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE tag "
            "("
            "handle VARCHAR(50) PRIMARY KEY NOT NULL, "
            "blob_data BLOB"
            ")"
        )
        # Secondary:
        self.dbapi.execute(
            "CREATE TABLE reference "
            "("
            "obj_handle VARCHAR(50), "
            "obj_class TEXT, "
            "ref_handle VARCHAR(50), "
            "ref_class TEXT"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE name_group "
            "("
            "name VARCHAR(50) PRIMARY KEY NOT NULL, "
            "grouping TEXT"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE metadata "
            "("
            "setting VARCHAR(50) PRIMARY KEY NOT NULL, "
            "value BLOB"
            ")"
        )
        self.dbapi.execute(
            "CREATE TABLE gender_stats "
            "("
            "given_name TEXT, "
            "female INTEGER, "
            "male INTEGER, "
            "unknown INTEGER"
            ")"
        )

        self._create_secondary_columns()

        ## Indices:
        self.dbapi.execute("CREATE INDEX person_gramps_id " "ON person(gramps_id)")
        self.dbapi.execute("CREATE INDEX person_surname " "ON person(surname)")
        self.dbapi.execute("CREATE INDEX person_given_name " "ON person(given_name)")
        self.dbapi.execute("CREATE INDEX source_title " "ON source(title)")
        self.dbapi.execute("CREATE INDEX source_gramps_id " "ON source(gramps_id)")
        self.dbapi.execute("CREATE INDEX citation_page " "ON citation(page)")
        self.dbapi.execute("CREATE INDEX citation_gramps_id " "ON citation(gramps_id)")
        self.dbapi.execute("CREATE INDEX media_desc " "ON media(desc)")
        self.dbapi.execute("CREATE INDEX media_gramps_id " "ON media(gramps_id)")
        self.dbapi.execute("CREATE INDEX place_title " "ON place(title)")
        self.dbapi.execute("CREATE INDEX place_enclosed_by " "ON place(enclosed_by)")
        self.dbapi.execute("CREATE INDEX place_gramps_id " "ON place(gramps_id)")
        self.dbapi.execute("CREATE INDEX tag_name " "ON tag(name)")
        self.dbapi.execute(
            "CREATE INDEX reference_ref_handle " "ON reference(ref_handle)"
        )
        self.dbapi.execute("CREATE INDEX family_gramps_id " "ON family(gramps_id)")
        self.dbapi.execute("CREATE INDEX event_gramps_id " "ON event(gramps_id)")
        self.dbapi.execute(
            "CREATE INDEX repository_gramps_id " "ON repository(gramps_id)"
        )
        self.dbapi.execute("CREATE INDEX note_gramps_id " "ON note(gramps_id)")
        self.dbapi.execute(
            "CREATE INDEX reference_obj_handle " "ON reference(obj_handle)"
        )

        self.dbapi.commit()

    def _close(self):
        self.dbapi.close()

    def _txn_begin(self):
        """
        Lowlevel interface to the backend transaction.
        Executes a db BEGIN;
        """
        if self.transaction == None:
            _LOG.debug("    DBAPI %s transaction begin", hex(id(self)))
            self.dbapi.begin()

    def _txn_commit(self):
        """
        Lowlevel interface to the backend transaction.
        Executes a db END;
        """
        if self.transaction == None:
            _LOG.debug("    DBAPI %s transaction commit", hex(id(self)))
            self.dbapi.commit()

    def _txn_abort(self):
        """
        Lowlevel interface to the backend transaction.
        Executes a db ROLLBACK;
        """
        if self.transaction == None:
            self.dbapi.rollback()

    def _collation(self, locale):
        """
        Get the adjusted collation if there is one, falling back on
        the locale.collation.
        """
        collation = self.dbapi.check_collation(locale)
        if collation == None:
            return locale.get_collation()
        return collation

    def transaction_begin(self, transaction):
        """
        Transactions are handled automatically by the db layer.
        """
        _LOG.debug(
            "    %sDBAPI %s transaction begin for '%s'",
            "Batch " if transaction.batch else "",
            hex(id(self)),
            transaction.get_description(),
        )
        if transaction.batch:
            # A batch transaction does not store the commits
            # Aborting the session completely will become impossible.
            self.abort_possible = False
        self.transaction = transaction
        self.dbapi.begin()
        return transaction

    def transaction_commit(self, txn):
        """
        Executed at the end of a transaction.
        """
        _LOG.debug(
            "    %sDBAPI %s transaction commit for '%s'",
            "Batch " if txn.batch else "",
            hex(id(self)),
            txn.get_description(),
        )

        action = {TXNADD: "-add", TXNUPD: "-update", TXNDEL: "-delete", None: "-delete"}
        self.dbapi.commit()
        if not txn.batch:
            # Now, emit signals:
            # do deletes and adds first
            for trans_type in [TXNDEL, TXNADD, TXNUPD]:
                for obj_type in range(11):
                    if obj_type != REFERENCE_KEY and (obj_type, trans_type) in txn:
                        if trans_type == TXNDEL:
                            handles = [
                                handle for (handle, data) in txn[(obj_type, trans_type)]
                            ]
                        else:
                            handles = [
                                handle
                                for (handle, data) in txn[(obj_type, trans_type)]
                                if (handle, None) not in txn[(obj_type, TXNDEL)]
                            ]
                        if handles:
                            signal = KEY_TO_NAME_MAP[obj_type] + action[trans_type]
                            self.emit(signal, (handles,))
        self.transaction = None
        msg = txn.get_description()
        self.undodb.commit(txn, msg)
        self._after_commit(txn)
        txn.clear()
        self.has_changed += 1  # Also gives commits since startup

    def transaction_abort(self, txn):
        """
        Executed after a batch operation abort.
        """
        self.dbapi.rollback()
        self.transaction = None
        txn.clear()
        txn.first = None
        txn.last = None
        self._after_commit(txn)

    def _get_metadata(self, key, default=[]):
        """
        Get an item from the database.

        Default is an empty list, which is a mutable and
        thus a bad default (pylint will complain).

        However, it is just used as a value, and not altered, so
        its use here is ok.
        """
        self.dbapi.execute("SELECT value FROM metadata WHERE setting = ?", [key])
        row = self.dbapi.fetchone()
        if row:
            return pickle.loads(row[0])
        elif default == []:
            return []
        else:
            return default

    def _set_metadata(self, key, value):
        """
        key: string
        value: item, will be serialized here
        """
        self._txn_begin()
        self.dbapi.execute("SELECT 1 FROM metadata WHERE setting = ?", [key])
        row = self.dbapi.fetchone()
        if row:
            self.dbapi.execute(
                "UPDATE metadata SET value = ? WHERE setting = ?",
                [pickle.dumps(value), key],
            )
        else:
            self.dbapi.execute(
                "INSERT INTO metadata (setting, value) VALUES (?, ?)",
                [key, pickle.dumps(value)],
            )
        self._txn_commit()

    def get_name_group_keys(self):
        """
        Return the defined names that have been assigned to a default grouping.
        """
        self.dbapi.execute("SELECT name, grouping FROM name_group " "ORDER BY name")
        rows = self.dbapi.fetchall()
        # not None test below fixes db corrupted by 11011 for export
        return [row[0] for row in rows if row[1] is not None]

    def get_name_group_mapping(self, key):
        """
        Return the default grouping name for a surname.
        """
        self.dbapi.execute("SELECT grouping FROM name_group WHERE name = ?", [key])
        row = self.dbapi.fetchone()
        if row and row[0] is not None:
            # not None test fixes db corrupted by 11011
            return row[0]
        else:
            return key

    def get_person_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Person in
        the database.

        :param sort_handles: If True, the list is sorted by surnames.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if sort_handles:
            self.dbapi.execute(
                "SELECT handle FROM person "
                "ORDER BY surname "
                'COLLATE "%s"' % self._collation(locale)
            )
        else:
            self.dbapi.execute("SELECT handle FROM person")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_family_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Family in
        the database.

        :param sort_handles: If True, the list is sorted by surnames.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if sort_handles:
            sql = (
                "SELECT family.handle "
                + "FROM family "
                + "LEFT JOIN person AS father "
                + "ON family.father_handle = father.handle "
                + "LEFT JOIN person AS mother "
                + "ON family.mother_handle = mother.handle "
                + "ORDER BY (CASE WHEN father.handle IS NULL "
                + "THEN mother.surname "
                + "ELSE father.surname "
                + "END), "
                + "(CASE WHEN family.handle IS NULL "
                + "THEN mother.given_name "
                + "ELSE father.given_name "
                + "END) "
                + 'COLLATE "%s"' % self._collation(locale)
            )
            self.dbapi.execute(sql)
        else:
            self.dbapi.execute("SELECT handle FROM family")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_event_handles(self):
        """
        Return a list of database handles, one handle for each Event in the
        database.
        """
        self.dbapi.execute("SELECT handle FROM event")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_citation_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Citation in
        the database.

        :param sort_handles: If True, the list is sorted by Citation title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if sort_handles:
            self.dbapi.execute(
                "SELECT handle FROM citation "
                "ORDER BY page "
                'COLLATE "%s"' % self._collation(locale)
            )
        else:
            self.dbapi.execute("SELECT handle FROM citation")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_source_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Source in
        the database.

        :param sort_handles: If True, the list is sorted by Source title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if sort_handles:
            self.dbapi.execute(
                "SELECT handle FROM source "
                "ORDER BY title "
                'COLLATE "%s"' % self._collation(locale)
            )
        else:
            self.dbapi.execute("SELECT handle from source")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_place_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Place in
        the database.

        :param sort_handles: If True, the list is sorted by Place title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if sort_handles:
            self.dbapi.execute(
                "SELECT handle FROM place "
                "ORDER BY title "
                'COLLATE "%s"' % self._collation(locale)
            )
        else:
            self.dbapi.execute("SELECT handle FROM place")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_repository_handles(self):
        """
        Return a list of database handles, one handle for each Repository in
        the database.
        """
        self.dbapi.execute("SELECT handle FROM repository")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_media_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Media in
        the database.

        :param sort_handles: If True, the list is sorted by title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if sort_handles:
            self.dbapi.execute(
                "SELECT handle FROM media "
                "ORDER BY desc "
                'COLLATE "%s"' % self._collation(locale)
            )
        else:
            self.dbapi.execute("SELECT handle FROM media")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_note_handles(self):
        """
        Return a list of database handles, one handle for each Note in the
        database.
        """
        self.dbapi.execute("SELECT handle FROM note")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_tag_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Tag in
        the database.

        :param sort_handles: If True, the list is sorted by Tag name.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if sort_handles:
            self.dbapi.execute(
                "SELECT handle FROM tag "
                "ORDER BY name "
                'COLLATE "%s"' % self._collation(locale)
            )
        else:
            self.dbapi.execute("SELECT handle FROM tag")
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def get_tag_from_name(self, name):
        """
        Find a Tag in the database from the passed Tag name.

        If no such Tag exists, None is returned.
        """
        self.dbapi.execute("SELECT blob_data FROM tag WHERE name = ?", [name])
        row = self.dbapi.fetchone()
        if row:
            return Tag.create(pickle.loads(row[0]))
        return None

    def _get_number_of(self, obj_key):
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT count(1) FROM %s" % table
        self.dbapi.execute(sql)
        row = self.dbapi.fetchone()
        return row[0]

    def has_name_group_key(self, key):
        """
        Return if a key exists in the name_group table.
        """
        self.dbapi.execute("SELECT grouping FROM name_group WHERE name = ?", [key])
        row = self.dbapi.fetchone()
        return row and row[0] is not None

    def set_name_group_mapping(self, name, grouping):
        """
        Set the default grouping name for a surname.
        """
        self._txn_begin()
        self.dbapi.execute("SELECT 1 FROM name_group WHERE name = ?", [name])
        row = self.dbapi.fetchone()
        if row and grouping is not None:
            self.dbapi.execute(
                "UPDATE name_group SET grouping=? " "WHERE name = ?", [grouping, name]
            )
        elif row and grouping is None:
            self.dbapi.execute("DELETE FROM name_group WHERE name = ?", [name])
        else:
            self.dbapi.execute(
                "INSERT INTO name_group (name, grouping) VALUES (?, ?)",
                [name, grouping],
            )
        self._txn_commit()
        if grouping is None:
            grouping = ""
        self.emit("person-groupname-rebuild", (name, grouping))

    def _commit_base(self, obj, obj_key, trans, change_time):
        """
        Commit the specified object to the database, storing the changes as
        part of the transaction.
        """
        old_data = None
        obj.change = int(change_time or time.time())
        table = KEY_TO_NAME_MAP[obj_key]

        if self._has_handle(obj_key, obj.handle):
            old_data = self._get_raw_data(obj_key, obj.handle)
            # update the object:
            sql = "UPDATE %s SET blob_data = ? WHERE handle = ?" % table
            self.dbapi.execute(sql, [pickle.dumps(obj.serialize()), obj.handle])
        else:
            # Insert the object:
            sql = ("INSERT INTO %s (handle, blob_data) VALUES (?, ?)") % table
            self.dbapi.execute(sql, [obj.handle, pickle.dumps(obj.serialize())])
        self._update_secondary_values(obj)
        self._update_backlinks(obj, trans)
        if not trans.batch:
            if old_data:
                trans.add(obj_key, TXNUPD, obj.handle, old_data, obj.serialize())
            else:
                trans.add(obj_key, TXNADD, obj.handle, None, obj.serialize())

        return old_data

    def _commit_raw(self, data, obj_key):
        """
        Commit a serialized primary object to the database, storing the
        changes as part of the transaction.
        """
        table = KEY_TO_NAME_MAP[obj_key]
        handle = data[0]

        if self._has_handle(obj_key, handle):
            # update the object:
            sql = "UPDATE %s SET blob_data = ? WHERE handle = ?" % table
            self.dbapi.execute(sql, [pickle.dumps(data), handle])
        else:
            # Insert the object:
            sql = ("INSERT INTO %s (handle, blob_data) VALUES (?, ?)") % table
            self.dbapi.execute(sql, [handle, pickle.dumps(data)])

        return

    def _update_backlinks(self, obj, transaction):
        if not transaction.batch:
            # Find existing references
            sql = (
                "SELECT ref_class, ref_handle " + "FROM reference WHERE obj_handle = ?"
            )
            self.dbapi.execute(sql, [obj.handle])
            existing_references = set(self.dbapi.fetchall())

            # Once we have the list of rows that already have a reference
            # we need to compare it with the list of objects that are
            # still references from the primary object.
            current_references = set(obj.get_referenced_handles_recursively())
            no_longer_required_references = existing_references.difference(
                current_references
            )
            new_references = current_references.difference(existing_references)

            # Delete the existing references
            self.dbapi.execute(
                "DELETE FROM reference WHERE obj_handle = ?", [obj.handle]
            )

            # Now, add the current ones
            for ref_class_name, ref_handle in current_references:
                sql = (
                    "INSERT INTO reference "
                    + "(obj_handle, obj_class, ref_handle, ref_class)"
                    + "VALUES(?, ?, ?, ?)"
                )
                self.dbapi.execute(
                    sql,
                    [obj.handle, obj.__class__.__name__, ref_handle, ref_class_name],
                )

            # Add new references to the transaction
            for ref_class_name, ref_handle in new_references:
                key = (obj.handle, ref_handle)
                data = (obj.handle, obj.__class__.__name__, ref_handle, ref_class_name)
                transaction.add(REFERENCE_KEY, TXNADD, key, None, data)

            # Add old references to the transaction
            for ref_class_name, ref_handle in no_longer_required_references:
                key = (obj.handle, ref_handle)
                old_data = (
                    obj.handle,
                    obj.__class__.__name__,
                    ref_handle,
                    ref_class_name,
                )
                transaction.add(REFERENCE_KEY, TXNDEL, key, old_data, None)
        else:  # batch mode
            current_references = set(obj.get_referenced_handles_recursively())

            # Delete the existing references
            self.dbapi.execute(
                "DELETE FROM reference WHERE obj_handle = ?", [obj.handle]
            )

            # Now, add the current ones
            for ref_class_name, ref_handle in current_references:
                sql = (
                    "INSERT INTO reference "
                    + "(obj_handle, obj_class, ref_handle, ref_class)"
                    + "VALUES(?, ?, ?, ?)"
                )
                self.dbapi.execute(
                    sql,
                    [obj.handle, obj.__class__.__name__, ref_handle, ref_class_name],
                )

    def _do_remove(self, handle, transaction, obj_key):
        if self.readonly or not handle:
            return
        if self._has_handle(obj_key, handle):
            data = self._get_raw_data(obj_key, handle)
            obj_class = KEY_TO_CLASS_MAP[obj_key]
            self._remove_backlinks(obj_class, handle, transaction)
            table = KEY_TO_NAME_MAP[obj_key]
            sql = "DELETE FROM %s WHERE handle = ?" % table
            self.dbapi.execute(sql, [handle])
            if not transaction.batch:
                transaction.add(obj_key, TXNDEL, handle, data, None)

    def _remove_backlinks(self, obj_class, obj_handle, transaction):
        """
        Removes all references from this object (backlinks).
        """
        # collect backlinks from this object for undo
        self.dbapi.execute(
            "SELECT ref_class, ref_handle " + "FROM reference WHERE obj_handle = ?",
            [obj_handle],
        )
        rows = self.dbapi.fetchall()
        # Now, delete backlinks from this object:
        self.dbapi.execute("DELETE FROM reference WHERE obj_handle = ?;", [obj_handle])
        # Add old references to the transaction
        if not transaction.batch:
            for ref_class_name, ref_handle in rows:
                key = (obj_handle, ref_handle)
                old_data = (obj_handle, obj_class, ref_handle, ref_class_name)
                transaction.add(REFERENCE_KEY, TXNDEL, key, old_data, None)

    def find_backlink_handles(self, handle, include_classes=None):
        """
        Find all objects that hold a reference to the object handle.

        Returns an interator over a list of (class_name, handle) tuples.

        :param handle: handle of the object to search for.
        :type handle: database handle
        :param include_classes: list of class names to include in the results.
            Default: None means include all classes.
        :type include_classes: list of class names

        Note that this is a generator function, it returns a iterator for
        use in loops. If you want a list of the results use::

            result_list = list(find_backlink_handles(handle))
        """
        self.dbapi.execute(
            "SELECT obj_class, obj_handle " "FROM reference " "WHERE ref_handle = ?",
            [handle],
        )
        rows = self.dbapi.fetchall()
        for row in rows:
            if (include_classes is None) or (row[0] in include_classes):
                yield (row[0], row[1])

    def find_initial_person(self):
        """
        Returns first person in the database
        """
        handle = self.get_default_handle()
        person = None
        if handle:
            person = self.get_person_from_handle(handle)
            if person:
                return person
        self.dbapi.execute("SELECT handle FROM person")
        row = self.dbapi.fetchone()
        if row:
            return self.get_person_from_handle(row[0])

    def _iter_handles(self, obj_key):
        """
        Return an iterator over handles in the database
        """
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT handle FROM %s" % table
        self.dbapi.execute(sql)
        rows = self.dbapi.fetchall()
        for row in rows:
            yield row[0]

    def _iter_raw_data(self, obj_key):
        """
        Return an iterator over raw data in the database.
        """
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT handle, blob_data FROM %s" % table
        with self.dbapi.cursor() as cursor:
            cursor.execute(sql)
            rows = cursor.fetchmany()
            while rows:
                for row in rows:
                    yield (row[0], pickle.loads(row[1]))
                rows = cursor.fetchmany()

    def _iter_raw_place_tree_data(self):
        """
        Return an iterator over raw data in the place hierarchy.
        """
        to_do = [""]
        sql = "SELECT handle, blob_data FROM place WHERE enclosed_by = ?"
        while to_do:
            handle = to_do.pop()
            self.dbapi.execute(sql, [handle])
            rows = self.dbapi.fetchall()
            for row in rows:
                to_do.append(row[0])
                yield (row[0], pickle.loads(row[1]))

    def reindex_reference_map(self, callback):
        """
        Reindex all primary records in the database.
        """
        self._txn_begin()
        self.dbapi.execute("DELETE FROM reference")
        total = 0
        for tbl in (
            "people",
            "families",
            "events",
            "places",
            "sources",
            "citations",
            "media",
            "repositories",
            "notes",
            "tags",
        ):
            total += self.method("get_number_of_%s", tbl)()
        UpdateCallback.__init__(self, callback)
        self.set_total(total)
        primary_table = (
            (self.get_person_cursor, Person),
            (self.get_family_cursor, Family),
            (self.get_event_cursor, Event),
            (self.get_place_cursor, Place),
            (self.get_source_cursor, Source),
            (self.get_citation_cursor, Citation),
            (self.get_media_cursor, Media),
            (self.get_repository_cursor, Repository),
            (self.get_note_cursor, Note),
            (self.get_tag_cursor, Tag),
        )
        # Now we use the functions and classes defined above
        # to loop through each of the primary object tables.
        for cursor_func, class_func in primary_table:
            logging.info("Rebuilding %s reference map", class_func.__name__)
            with cursor_func() as cursor:
                for found_handle, val in cursor:
                    obj = class_func.create(val)
                    references = set(obj.get_referenced_handles_recursively())
                    # handle addition of new references
                    for ref_class_name, ref_handle in references:
                        self.dbapi.execute(
                            "INSERT INTO reference "
                            "(obj_handle, obj_class, ref_handle, ref_class) "
                            "VALUES (?, ?, ?, ?)",
                            [
                                obj.handle,
                                obj.__class__.__name__,
                                ref_handle,
                                ref_class_name,
                            ],
                        )
                    self.update()
        self._txn_commit()

    def rebuild_secondary(self, callback=None):
        """
        Rebuild secondary indices
        """
        if self.readonly:
            return

        total = 0
        for tbl in (
            "people",
            "families",
            "events",
            "places",
            "sources",
            "citations",
            "media",
            "repositories",
            "notes",
            "tags",
        ):
            total += self.method("get_number_of_%s", tbl)()
        UpdateCallback.__init__(self, callback)
        self.set_total(total)

        # First, expand blob to individual fields:
        self._txn_begin()
        for obj_type in (
            "Person",
            "Family",
            "Event",
            "Place",
            "Repository",
            "Source",
            "Citation",
            "Media",
            "Note",
            "Tag",
        ):
            for handle in self.method("get_%s_handles", obj_type)():
                obj = self.method("get_%s_from_handle", obj_type)(handle)
                self._update_secondary_values(obj)
                self.update()
        self._txn_commit()

        # Next, rebuild stats:
        gstats = self.get_gender_stats()
        self.genderStats = GenderStats(gstats)

    def _has_handle(self, obj_key, handle):
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT 1 FROM %s WHERE handle = ?" % table
        self.dbapi.execute(sql, [handle])
        return self.dbapi.fetchone() is not None

    def _has_gramps_id(self, obj_key, gramps_id):
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT 1 FROM %s WHERE gramps_id = ?" % table
        self.dbapi.execute(sql, [gramps_id])
        return self.dbapi.fetchone() != None

    def _get_gramps_ids(self, obj_key):
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT gramps_id FROM %s" % table
        self.dbapi.execute(sql)
        rows = self.dbapi.fetchall()
        return [row[0] for row in rows]

    def _get_raw_data(self, obj_key, handle):
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT blob_data FROM %s WHERE handle = ?" % table
        self.dbapi.execute(sql, [handle])
        row = self.dbapi.fetchone()
        if row:
            return pickle.loads(row[0])

    def _get_raw_from_id_data(self, obj_key, gramps_id):
        table = KEY_TO_NAME_MAP[obj_key]
        sql = "SELECT blob_data FROM %s WHERE gramps_id = ?" % table
        self.dbapi.execute(sql, [gramps_id])
        row = self.dbapi.fetchone()
        if row:
            return pickle.loads(row[0])

    def get_gender_stats(self):
        """
        Returns a dictionary of
        {given_name: (male_count, female_count, unknown_count)}
        """
        self.dbapi.execute(
            "SELECT given_name, female, male, unknown " "FROM gender_stats"
        )
        gstats = {}
        for row in self.dbapi.fetchall():
            gstats[row[0]] = (row[1], row[2], row[3])
        return gstats

    def save_gender_stats(self, gstats):
        self._txn_begin()
        self.dbapi.execute("DELETE FROM gender_stats")
        for key in gstats.stats:
            female, male, unknown = gstats.stats[key]
            self.dbapi.execute(
                "INSERT INTO gender_stats "
                "(given_name, female, male, unknown) "
                "VALUES (?, ?, ?, ?)",
                [key, female, male, unknown],
            )
        self._txn_commit()

    def undo_reference(self, data, handle):
        """
        Helper method to undo a reference map entry
        """
        if data is None:
            sql = "DELETE FROM reference " + "WHERE obj_handle = ? AND ref_handle = ?"
            self.dbapi.execute(sql, [handle[0], handle[1]])
        else:
            sql = (
                "INSERT INTO reference "
                + "(obj_handle, obj_class, ref_handle, ref_class) "
                + "VALUES(?, ?, ?, ?)"
            )
            self.dbapi.execute(sql, data)

    def undo_data(self, data, handle, obj_key):
        """
        Helper method to undo/redo the changes made
        """
        cls = KEY_TO_CLASS_MAP[obj_key]
        table = cls.lower()
        if data is None:
            sql = "DELETE FROM %s WHERE handle = ?" % table
            self.dbapi.execute(sql, [handle])
        else:
            if self._has_handle(obj_key, handle):
                sql = "UPDATE %s SET blob_data = ? WHERE handle = ?" % table
                self.dbapi.execute(sql, [pickle.dumps(data), handle])
            else:
                sql = "INSERT INTO %s (handle, blob_data) VALUES (?, ?)" % table
                self.dbapi.execute(sql, [handle, pickle.dumps(data)])
            obj = self._get_table_func(cls)["class_func"].create(data)
            self._update_secondary_values(obj)

    def get_surname_list(self):
        """
        Return the list of locale-sorted surnames contained in the database.
        """
        self.dbapi.execute("SELECT DISTINCT surname " "FROM person " "ORDER BY surname")
        surname_list = []
        for row in self.dbapi.fetchall():
            surname_list.append(row[0])
        return surname_list

    def _sql_type(self, schema_type, max_length):
        """
        Given a schema type, return the SQL type for
        a new column.
        """
        if schema_type == "string":
            if max_length:
                return "VARCHAR(%s)" % max_length
            else:
                return "TEXT"
        elif schema_type in ["boolean", "integer"]:
            return "INTEGER"
        elif schema_type == "number":
            return "REAL"
        else:
            return "BLOB"

    def _create_secondary_columns(self):
        """
        Create secondary columns.
        """
        LOG.debug("Creating secondary columns...")
        for cls in (
            Person,
            Family,
            Event,
            Place,
            Repository,
            Source,
            Citation,
            Media,
            Note,
            Tag,
        ):
            table_name = cls.__name__.lower()
            for field, schema_type, max_length in cls.get_secondary_fields():
                if field != "handle":
                    sql_type = self._sql_type(schema_type, max_length)
                    self.dbapi.execute(
                        "ALTER TABLE %s ADD COLUMN %s %s"
                        % (table_name, field, sql_type)
                    )

    def _update_secondary_values(self, obj):
        """
        Given a primary object update its secondary field values
        in the database.
        Does not commit.
        """
        table = obj.__class__.__name__
        fields = [field[0] for field in obj.get_secondary_fields()]
        sets = []
        values = []
        for field in fields:
            sets.append("%s = ?" % field)
            values.append(getattr(obj, field))

        # Derived fields
        if table == "Person":
            given_name, surname = self._get_person_data(obj)
            sets.append("given_name = ?")
            values.append(given_name)
            sets.append("surname = ?")
            values.append(surname)
        if table == "Place":
            handle = self._get_place_data(obj)
            sets.append("enclosed_by = ?")
            values.append(handle)

        if len(values) > 0:
            table_name = table.lower()
            self.dbapi.execute(
                "UPDATE %s SET %s where handle = ?" % (table_name, ", ".join(sets)),
                self._sql_cast_list(values) + [obj.handle],
            )

    def _sql_cast_list(self, values):
        """
        Given a list of field names and values, return the values
        in the appropriate type.
        """
        return [v if not isinstance(v, bool) else int(v) for v in values]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2019       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
Provides a simple histogram widget for use in gramplets.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import math

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Pango, PangoCairo

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from ...gen.const import GRAMPS_LOCALE as glocale


class Histogram(Gtk.DrawingArea):
    """
    A simple histogram widget for use in gramplets.
    """

    __gsignals__ = {"clicked": (GObject.SignalFlags.RUN_FIRST, None, (int,))}

    def __init__(self):
        Gtk.DrawingArea.__init__(self)

        self.add_events(
            Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
        )
        self.connect("motion-notify-event", self.on_pointer_motion)
        self.connect("button-press-event", self.on_button_press)

        self.title = ""
        self.bucket_axis = ""
        self.value_axis = ""
        self.grid_lines = True
        self.data = []
        self.labels = []
        self.tooltip = ""
        self.highlight = None
        self.__bars = None
        self.__active = -1

    def set_title(self, title):
        """
        Set the main chart title.
        @param title: The main chart title.
        @type title: str
        """
        self.title = title

    def set_bucket_axis(self, bucket_axis):
        """
        Set the bucket axis label.
        @param bucket_axis: The bucket axis label.
        @type bucket_axis: str
        """
        self.bucket_axis = bucket_axis

    def set_value_axis(self, value_axis):
        """
        Set the value axis label.
        @param bucket_axis: The value axis label.
        @type bucket_axis: str
        """
        self.value_axis = value_axis

    def set_grid_lines(self, grid_lines):
        """
        Specify if grid lines should be displayed.
        @param grid_lines: True if grid lines should be displayed.
        @type grid_lines: bool
        """
        self.grid_lines = grid_lines

    def set_values(self, data):
        """
        Set the chart values.
        @param data: A list of values, one for each bucket.
        @type data: list
        """
        self.data = data

    def set_labels(self, labels):
        """
        Set the labels on the bucket axis.
        @param labels: A list of labels, one for each bucket.
        @type labels: list
        """
        self.labels = labels

    def set_tooltip(self, tooltip):
        """
        Set the tooltip to display on bars.  If the string contains a "%d"
        substitution variable it will be replaced with the value that the
        bar represents.
        @param labels: A tooltip.
        @type labels: str
        """
        self.tooltip = tooltip

    def set_highlight(self, highlight):
        """
        Specify the bars to hightlight.
        @param labels: A list of bucket numbers.
        @type labels: list
        """
        self.highlight = highlight

    def do_draw(self, cr):
        """
        A custom draw method for this widget.
        @param cr: A cairo context.
        @type cr: cairo.Context
        """
        allocation = self.get_allocation()
        context = self.get_style_context()
        fg_color = context.get_color(context.get_state())
        cr.set_source_rgba(*fg_color)

        # Title
        layout = self.create_pango_layout(self.title)
        width, height = layout.get_pixel_size()
        offset = height + 5

        # Labels
        label_width = 0
        for i in range(len(self.labels)):
            layout = self.create_pango_layout(self.labels[i])
            width, height = layout.get_pixel_size()
            if width > label_width:
                label_width = width
            cr.move_to(0, i * height + offset)
            PangoCairo.show_layout(cr, layout)

        layout = self.create_pango_layout(self.bucket_axis)
        width, height = layout.get_pixel_size()
        if width > label_width:
            label_width = width
        label_width += 5
        cr.move_to((label_width - width) / 2, 0)
        PangoCairo.show_layout(cr, layout)

        # Values
        percent_width = 0
        total = sum(self.data)
        for i in range(len(self.data)):
            if total > 0:
                percent = glocale.format_string("%.2f", self.data[i] / total * 100)
            else:
                percent = ""
            layout = self.create_pango_layout(percent)
            width, height = layout.get_pixel_size()
            if width > percent_width:
                percent_width = width
            cr.move_to(allocation.width - width, i * height + offset)
            PangoCairo.show_layout(cr, layout)

        layout = self.create_pango_layout(self.value_axis)
        width, height = layout.get_pixel_size()
        if width > percent_width:
            percent_width = width
        percent_width += 5
        cr.move_to(allocation.width - (percent_width + width) / 2, 0)
        PangoCairo.show_layout(cr, layout)

        chart_width = allocation.width - label_width - percent_width
        spacing = 2

        # Title
        layout = self.create_pango_layout(self.title)
        layout.set_ellipsize(Pango.EllipsizeMode.END)
        layout.set_width((chart_width - 10) * Pango.SCALE)
        cr.move_to(label_width + 5, 0)
        PangoCairo.show_layout(cr, layout)

        # Border
        cr.move_to(0, offset)
        cr.line_to(allocation.width, offset)
        cr.stroke()

        bottom = len(self.data) * height + (2 * spacing) + offset
        cr.move_to(0, bottom)
        cr.line_to(allocation.width, bottom)
        cr.stroke()

        cr.move_to(label_width, 0)
        cr.line_to(label_width, bottom)
        cr.stroke()

        cr.move_to(allocation.width - percent_width, 0)
        cr.line_to(allocation.width - percent_width, bottom)
        cr.stroke()

        # Ticks and grid lines
        tick_step, maximum = self.__get_tick_step(chart_width)
        count = 0
        while count <= maximum:
            # draw tick
            tick_pos = label_width + chart_width * count / maximum
            cr.move_to(tick_pos, bottom)
            cr.line_to(tick_pos, bottom + 5)
            cr.stroke()
            # draw grid line
            if self.grid_lines:
                cr.set_dash([1, 2])
                cr.move_to(tick_pos, bottom)
                cr.line_to(tick_pos, (2 * spacing) + offset)
                cr.stroke()
                cr.set_dash([])
            layout = self.create_pango_layout("%d" % count)
            width, height = layout.get_pixel_size()
            cr.move_to(tick_pos - (width / 2), bottom + 5)
            PangoCairo.show_layout(cr, layout)
            count += tick_step

        # Bars
        cr.set_line_width(1)
        bar_size = height - (2 * spacing)
        self.__bars = []
        for i in range(len(self.labels)):
            cr.rectangle(
                label_width,
                i * height + (2 * spacing) + offset,
                chart_width * self.data[i] / maximum,
                bar_size,
            )
            self.__bars.append(
                [
                    label_width,
                    i * height + (2 * spacing) + offset,
                    chart_width * self.data[i] / maximum,
                    bar_size,
                ]
            )
            if i in self.highlight:
                if self.__active == i:
                    cr.set_source_rgba(1, 0.7, 0, 1)
                else:
                    cr.set_source_rgba(1, 0.5, 0, 1)
            else:
                if self.__active == i:
                    cr.set_source_rgba(0.7, 0.7, 1, 1)
                else:
                    cr.set_source_rgba(0.5, 0.5, 1, 1)

            cr.fill_preserve()
            cr.set_source_rgba(*fg_color)
            cr.stroke()

        self.set_size_request(-1, bottom + height + 5)

    def __get_tick_step(self, chart_width):
        """
        A method used to calculate the value axis scale and label spacing.
        @param chart_width: The chart size in pixels.
        @type chart_width: int
        """
        max_data = max(self.data)
        if max_data == 0:
            return 1, 1
        digits = int(math.log10(max_data)) + 1
        ticks = chart_width / (digits * 10 * 3)
        approx_step = max_data / ticks
        if approx_step < 1:
            approx_step = 1
        multiplier = 10 ** int(math.log10(approx_step))
        intervals = [1, 2, 5, 10]
        for interval in intervals:
            if interval >= approx_step / multiplier:
                break
        step = interval * multiplier
        max_value = (int(max_data / step) + 1) * step
        return step, max_value

    def on_pointer_motion(self, _dummy, event):
        """
        Called when the pointer is moved.
        @param _dummy: This widget.  Unused.
        @type _dummy: Gtk.Widget
        @param event: An event.
        @type event: Gdk.Event
        """
        if self.__bars is None:
            return False
        active = -1
        for i, bar in enumerate(self.__bars):
            if (
                event.x > bar[0]
                and event.x < bar[0] + bar[2]
                and event.y > bar[1]
                and event.y < bar[1] + bar[3]
            ):
                active = i
        if self.__active != active:
            self.__active = active
            self.queue_draw()
            if active == -1:
                self.set_tooltip_text("")
            else:
                if "%d" in self.tooltip:
                    self.set_tooltip_text(self.tooltip % self.data[active])
                else:
                    self.set_tooltip_text(self.tooltip)
        return False

    def on_button_press(self, _dummy, event):
        """
        Called when a mouse button is clicked.
        @param _dummy: This widget.  Unused.
        @type _dummy: Gtk.Widget
        @param event: An event.
        @type event: Gdk.Event
        """
        if (
            event.button == 1
            and event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS
            and self.__active != -1
        ):
            self.emit("clicked", self.__active)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2018       Robin van der Vliet
# Copyright (C) 2020       Jan Sparreboom
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
Specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------
#
# Levels
#
# -------------------------------------------------------------------------

_ordinal_level = [
    "",
    "eerste",
    "tweede",
    "derde",
    "vierde",
    "vijfde",
    "zesde",
    "zevende",
    "achtste",
    "negende",
    "tiende",
    "elfde",
    "twaalfde",
    "dertiende",
    "veertiende",
    "vijftiende",
    "zestiende",
    "zeventiende",
    "achttiende",
    "negentiende",
    "twintigste",
    "eenentwintigste",
    "tweeëntwintigste",
    "drieëntwintigste",
    "vierentwintigste",
    "vijfentwintigste",
    "zesentwintigste",
    "zevenentwintigste",
    "achtentwintigste",
    "negenentwintigste",
    "dertigste",
    "eenendertigste",
    "tweeëndertigste",
    "drieëndertigste",
    "vierendertigste",
    "vijfendertigste",
    "zesendertigste",
    "zevenendertigste",
    "achtendertigste",
    "negenendertigste",
    "veertigste",
    "eenenveertigste",
    "tweeënveertigste",
    "drieënveertigste",
    "vierenveertigste",
    "vijfenveertigste",
    "zesenveertigste",
    "zevenenveertigste",
    "achtenveertigste",
    "negenenveertigste",
    "vijftigste",
]

_removed_level = [
    "",
    "",
    "groot",
    "overgroot",
    "betovergroot",
    "oud",
    "oudgroot",
    "oudovergroot",
    "oudbetovergroot",
    "stam",
    "stamgroot",
    "stamovergroot",
    "stambetovergroot",
    "stamoud",
    "stamoudgroot",
    "stamoudovergroot",
    "stamoudbetovergroot",
    "edel",
    "edelgroot",
    "edelovergroot",
    "edelbetovergroot",
    "edeloud",
    "edeloudgroot",
    "edeloudovergroot",
    "edeloudbetovergroot",
    "edelstam",
    "edelstamgroot",
    "edelstamovergroot",
    "edelstambetovergroot",
    "edelstamoud",
    "edelstamoudgroot",
    "edelstamoudovergroot",
    "edelstamoudbetovergroot",
    "voor",
    "voorgroot",
    "voorovergroot",
    "voorbetovergroot",
    "vooroud",
    "vooroudgroot",
    "vooroudovergroot",
    "vooroudbetovergroot",
    "voorstam",
    "voorstamgroot",
    "voorstamovergroot",
    "voorstambetovergroot",
    "voorstamoud",
    "voorstamoudgroot",
    "voorstamoudovergroot",
    "voorstamoudbetovergroot",
    "vooredel",
]

_child_level = ["", "", "klein", "achterklein", "achterachterklein"]

_nibling_level = ["", "", "achter", "achterachter", "achterachterachter"]

_parents_level = [
    "",
    "ouders",
    "grootouders",
    "overgrootouders",
    "betovergrootouders",
    "overgrootouders 5e graad",
    "overgrootouders 6e graad",
    "overgrootouders 7e graad",
    "overgrootouders 8e graad",
    "overgrootouders 9e graad",
    "overgrootouders 10e graad",
    "overgrootouders 11e graad",
    "overgrootouders 12e graad",
    "overgrootouders 13e graad",
    "overgrootouders 14e graad",
    "overgrootouders 15e graad",
    "overgrootouders 16e graad",
    "overgrootouders 17e graad",
    "overgrootouders 18e graad",
    "overgrootouders 19e graad",
    "overgrootouders 20e graad",
    "overgrootouders 21e graad",
    "overgrootouders 22e graad",
    "overgrootouders 23e graad",
    "overgrootouders 24e graad",
    "overgrootouders 25e graad",
    "overgrootouders 26e graad",
    "overgrootouders 27e graad",
    "overgrootouders 28e graad",
    "overgrootouders 29e graad",
    "overgrootouders 30e graad",
    "overgrootouders 31e graad",
    "overgrootouders 32e graad",
    "overgrootouders 33e graad",
    "overgrootouders 34e graad",
    "overgrootouders 35e graad",
    "overgrootouders 36e graad",
    "overgrootouders 37e graad",
    "overgrootouders 38e graad",
    "overgrootouders 39e graad",
    "overgrootouders 40e graad",
    "overgrootouders 41e graad",
    "overgrootouders 42e graad",
    "overgrootouders 43e graad",
    "overgrootouders 44e graad",
    "overgrootouders 45e graad",
    "overgrootouders 46e graad",
    "overgrootouders 47e graad",
    "overgrootouders 48e graad",
    "overgrootouders 49e graad",
    "overgrootouders 50e graad",
]

_siblings_level = [
    "",
    "broers en zussen",
    "ooms en tantes",
    "oudooms en -tantes",
    "overoudooms en -tantes",
    "overoudooms en -tantes 5e graad",
    "overoudooms en -tantes 6e graad",
    "overoudooms en -tantes 7e graad",
    "overoudooms en -tantes 8e graad",
    "overoudooms en -tantes 9e graad",
    "overoudooms en -tantes 10e graad",
    "overoudooms en -tantes 11e graad",
    "overoudooms en -tantes 12e graad",
    "overoudooms en -tantes 13e graad",
    "overoudooms en -tantes 14e graad",
    "overoudooms en -tantes 15e graad",
    "overoudooms en -tantes 16e graad",
    "overoudooms en -tantes 17e graad",
    "overoudooms en -tantes 18e graad",
    "overoudooms en -tantes 19e graad",
    "overoudooms en -tantes 20e graad",
    "overoudooms en -tantes 21e graad",
    "overoudooms en -tantes 22e graad",
    "overoudooms en -tantes 23e graad",
    "overoudooms en -tantes 24e graad",
    "overoudooms en -tantes 25e graad",
    "overoudooms en -tantes 26e graad",
    "overoudooms en -tantes 27e graad",
    "overoudooms en -tantes 28e graad",
    "overoudooms en -tantes 29e graad",
    "overoudooms en -tantes 30e graad",
    "overoudooms en -tantes 41e graad",
    "overoudooms en -tantes 42e graad",
    "overoudooms en -tantes 43e graad",
    "overoudooms en -tantes 44e graad",
    "overoudooms en -tantes 45e graad",
    "overoudooms en -tantes 46e graad",
    "overoudooms en -tantes 47e graad",
    "overoudooms en -tantes 48e graad",
    "overoudooms en -tantes 49e graad",
    "overoudooms en -tantes 50e graad",
]

_children_level = [
    "",
    "kinderen",
    "kleinkinderen",
    "achterkleinkinderen",
    "betachterkleinkinderen",
    "kleinkinderen 5e graad",
    "kleinkinderen 6e graad",
    "kleinkinderen 7e graad",
    "kleinkinderen 8e graad",
    "kleinkinderen 9e graad",
    "kleinkinderen 10e graad",
    "kleinkinderen 11e graad",
    "kleinkinderen 12e graad",
    "kleinkinderen 13e graad",
    "kleinkinderen 14e graad",
    "kleinkinderen 15e graad",
    "kleinkinderen 16e graad",
    "kleinkinderen 17e graad",
    "kleinkinderen 18e graad",
    "kleinkinderen 19e graad",
    "kleinkinderen 20e graad",
    "kleinkinderen 21e graad",
    "kleinkinderen 22e graad",
    "kleinkinderen 23e graad",
    "kleinkinderen 24e graad",
    "kleinkinderen 25e graad",
    "kleinkinderen 26e graad",
    "kleinkinderen 27e graad",
    "kleinkinderen 28e graad",
    "kleinkinderen 29e graad",
    "kleinkinderen 30e graad",
    "kleinkinderen 31e graad",
    "kleinkinderen 32e graad",
    "kleinkinderen 33e graad",
    "kleinkinderen 34e graad",
    "kleinkinderen 35e graad",
    "kleinkinderen 36e graad",
    "kleinkinderen 37e graad",
    "kleinkinderen 38e graad",
    "kleinkinderen 39e graad",
    "kleinkinderen 40e graad",
    "kleinkinderen 41e graad",
    "kleinkinderen 42e graad",
    "kleinkinderen 43e graad",
    "kleinkinderen 44e graad",
    "kleinkinderen 45e graad",
    "kleinkinderen 46e graad",
    "kleinkinderen 47e graad",
    "kleinkinderen 48e graad",
    "kleinkinderen 49e graad",
    "kleinkinderen 50e graad",
]

_nephews_nieces_level = [
    "",
    "broers en zussen",
    "neven en nichten",
    "achterneven en -nichten",
    "achterneven en -nichten 4e graad",
    "achterneven en -nichten 5e graad",
    "achterneven en -nichten 6e graad",
    "achterneven en -nichten 7e graad",
    "achterneven en -nichten 8e graad",
    "achterneven en -nichten 9e graad",
    "achterneven en -nichten 10e graad",
    "achterneven en -nichten 11e graad",
    "achterneven en -nichten 12e graad",
    "achterneven en -nichten 13e graad",
    "achterneven en -nichten 14e graad",
    "achterneven en -nichten 15e graad",
    "achterneven en -nichten 16e graad",
    "achterneven en -nichten 17e graad",
    "achterneven en -nichten 18e graad",
    "achterneven en -nichten 19e graad",
    "achterneven en -nichten 20e graad",
    "achterneven en -nichten 21e graad",
    "achterneven en -nichten 22e graad",
    "achterneven en -nichten 23e graad",
    "achterneven en -nichten 24e graad",
    "achterneven en -nichten 25e graad",
    "achterneven en -nichten 26e graad",
    "achterneven en -nichten 27e graad",
    "achterneven en -nichten 28e graad",
    "achterneven en -nichten 29e graad",
    "achterneven en -nichten 30e graad",
    "achterneven en -nichten 31e graad",
    "achterneven en -nichten 32e graad",
    "achterneven en -nichten 33e graad",
    "achterneven en -nichten 34e graad",
    "achterneven en -nichten 35e graad",
    "achterneven en -nichten 36e graad",
    "achterneven en -nichten 37e graad",
    "achterneven en -nichten 38e graad",
    "achterneven en -nichten 39e graad",
    "achterneven en -nichten 40e graad",
    "achterneven en -nichten 41e graad",
    "achterneven en -nichten 42e graad",
    "achterneven en -nichten 43e graad",
    "achterneven en -nichten 44e graad",
    "achterneven en -nichten 45e graad",
    "achterneven en -nichten 46e graad",
    "achterneven en -nichten 47e graad",
    "achterneven en -nichten 48e graad",
    "achterneven en -nichten 49e graad",
    "achterneven en -nichten 50e graad",
]

# -------------------------------------------------------------------------
#
# Relationship calculator Dutch version
#
# -------------------------------------------------------------------------


class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    # sibling strings
    STEP = "stief"
    HALF = "half"

    INLAW = "aangetrouwde "

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_parents(self, level):
        if level > len(_removed_level) - 1:
            return "verre voorouders (%d generaties)" % level
        elif level > 4:
            return "%souders (%d generaties)" % (_removed_level[level], level)
        else:
            return "%souders" % _removed_level[level]

    def _get_father(self, level, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if level > len(_removed_level) - 1:
            return "verre %s%svoorvader (%d generaties)" % (inlaw, step, level)
        elif level > 4:
            return "%s%s%svader (%d generaties)" % (
                inlaw,
                step,
                _removed_level[level],
                level,
            )
        else:
            return "%s%s%svader" % (inlaw, step, _removed_level[level])

    def _get_son(self, level, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if level > len(_child_level) - 1:
            return "verre %s%sachterkleinzoon (%d generaties)" % (inlaw, step, level)
        else:
            return "%s%s%szoon" % (inlaw, step, _child_level[level])

    def _get_mother(self, level, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if level > len(_removed_level) - 1:
            return "verre %s%svoormoeder (%d generaties)" % (inlaw, step, level)
        elif level > 4:
            return "%s%s%smoeder (%d generaties)" % (
                inlaw,
                step,
                _removed_level[level],
                level,
            )
        else:
            return "%s%s%smoeder" % (inlaw, step, _removed_level[level])

    def _get_daughter(self, level, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if level > len(_child_level) - 1:
            return "verre %s%sachterkleindochter (%d generaties)" % (inlaw, step, level)
        else:
            return "%s%s%sdochter" % (inlaw, step, _child_level[level])

    def _get_parent_unknown(self, level, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if level > len(_removed_level) - 1:
            return "verre %s%svoorouder (%d generaties)" % (inlaw, step, level)
        elif level > 4:
            return "%s%s%souder (%d generaties)" % (
                inlaw,
                step,
                _removed_level[level],
                level,
            )
        else:
            return "%s%s%souder" % (inlaw, step, _removed_level[level])

    def _get_child_unknown(self, level, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if inlaw == "aangetrouwde ":
            # The word "kind" is grammatically neuter, so it has a different adjective.
            inlaw = "aangetrouwd "

        if level > len(_child_level) - 1:
            return "ver %s%sachterkleinkind (%d generaties)" % (inlaw, step, level)
        else:
            return "%s%s%skind" % (inlaw, step, _child_level[level])

    def _get_aunt(self, level, removed, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if removed == 1 and level < len(_removed_level):
            if level > 4:
                return "%s%s%stante (%d generaties)" % (
                    inlaw,
                    step,
                    _removed_level[level],
                    level,
                )
            else:
                return "%s%s%stante" % (inlaw, step, _removed_level[level])
        elif removed == 1:
            return "verre %s%stante (%d generaties)" % (inlaw, step, level)
        elif level > len(_removed_level) - 1 and removed > len(_ordinal_level) - 1:
            return "verre %s%stante (%d generaties, %d graden)" % (
                inlaw,
                step,
                level,
                removed,
            )
        elif level > len(_removed_level) - 1:
            return "verre %s%stante van de %s graad (%d generaties)" % (
                inlaw,
                step,
                _ordinal_level[removed],
                level,
            )
        else:
            if level > 4:
                return (
                    "%s%s%stante (%d generaties)"
                    % (inlaw, step, _removed_level[level], level)
                    + " "
                    + _ordinal_level[removed]
                    + " graad"
                )
            else:
                return (
                    "%s%s%stante" % (inlaw, step, _removed_level[level])
                    + " "
                    + _ordinal_level[removed]
                    + " graad"
                )

    def _get_uncle(self, level, removed, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if removed == 1 and level < len(_removed_level):
            if level > 4:
                return "%s%s%soom (%d generaties)" % (
                    inlaw,
                    step,
                    _removed_level[level],
                    level,
                )
            else:
                return "%s%s%soom" % (inlaw, step, _removed_level[level])
        elif removed == 1:
            return "verre %s%soom (%d generaties)" % (inlaw, step, level)
        elif level > len(_removed_level) - 1 and removed > len(_ordinal_level) - 1:
            return "verre %s%soom (%d generaties, %d graden)" % (
                inlaw,
                step,
                level,
                removed,
            )
        elif level > len(_removed_level) - 1:
            return "verre %s%soom van de %s graad (%d generaties)" % (
                inlaw,
                step,
                _ordinal_level[removed],
                level,
            )
        else:
            if level > 4:
                return (
                    "%s%s%soom (%d generaties)"
                    % (inlaw, step, _removed_level[level], level)
                    + " "
                    + _ordinal_level[removed]
                    + " graad"
                )
            else:
                return (
                    "%s%s%soom" % (inlaw, step, _removed_level[level])
                    + " "
                    + _ordinal_level[removed]
                    + " graad"
                )

    def _get_sibling(self, level, step="", inlaw=""):
        """overwrite of English method to return unknown gender sibling"""
        assert level == 1
        return (
            self._get_male_cousin(0, step=step, inlaw=inlaw)
            + " of "
            + self._get_female_cousin(0, step=step, inlaw=inlaw)
        )

    def _get_nephew(self, level, removed=1, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if removed == 1 and level < len(_nibling_level):
            return "%s%s%sneef" % (inlaw, step, _nibling_level[level])
        elif removed == 1:
            return "verre %s%sneef (%d generaties)" % (inlaw, step, level)
        elif level > len(_nibling_level) - 1 and removed > len(_ordinal_level) - 1:
            return "verre %s%sneef (%d generaties, %d graden)" % (
                inlaw,
                step,
                level,
                removed,
            )
        elif level > len(_nibling_level) - 1:
            return "verre %s%sneef van de %s graad (%d generaties)" % (
                inlaw,
                step,
                _ordinal_level[removed],
                level,
            )
        else:
            return (
                "%s%s%sneef" % (inlaw, step, _nibling_level[level])
                + " "
                + _ordinal_level[removed]
                + " graad"
            )

    def _get_niece(self, level, removed=1, step="", inlaw=""):
        """Internal Dutch method to create relation string"""
        if removed == 1 and level < len(_nibling_level):
            return "%s%s%snicht" % (inlaw, step, _nibling_level[level])
        elif removed == 1:
            return "verre %s%snicht (%d generaties)" % (inlaw, step, level)
        elif level > len(_nibling_level) - 1 and removed > len(_ordinal_level) - 1:
            return "verre %s%snicht (%d generaties, %d graden)" % (
                inlaw,
                step,
                level,
                removed,
            )
        elif level > len(_nibling_level) - 1:
            return "verre %s%snicht van de %s graad (%d generaties)" % (
                inlaw,
                step,
                _ordinal_level[removed],
                level,
            )
        else:
            return (
                "%s%s%snicht" % (inlaw, step, _nibling_level[level])
                + " "
                + _ordinal_level[removed]
                + " graad"
            )

    def _get_male_cousin(self, removed, step="", inlaw=""):
        """Specific Dutch thing, the nieces/nephews on same level are called
        going sideways in a branch as the nieces/newphews going downward
        from your brother/sisters. This used to be called "kozijn"
        """
        removed -= 1
        if removed > len(_ordinal_level) - 1:
            return "verre %s%sneef (kozijn, %d graden)" % (inlaw, step, removed)
        elif removed == 0:
            return "%s%sbroer" % (inlaw, step)
        else:
            return (
                "%s%sneef (kozijn)" % (inlaw, step)
                + " "
                + _ordinal_level[removed]
                + " graad"
            )

    def _get_female_cousin(self, removed, step="", inlaw=""):
        """Specific Dutch thing, the nieces/nephews on same level are called
        going sideways in a branch as the nieces/newphews going downward
        from your brother/sisters. This used to be called "kozijn"
        """
        removed -= 1
        if removed > len(_ordinal_level) - 1:
            return "verre %s%snicht (kozijn, %d graden)" % (inlaw, step, removed)
        elif removed == 0:
            return "%s%szus" % (inlaw, step)
        else:
            return (
                "%s%snicht (kozijn)" % (inlaw, step)
                + " "
                + _ordinal_level[removed]
                + " graad"
            )

    # NIEUW
    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Provide a string that describes the relationsip between a person, and
        a group of people with the same relationship. E.g. "grandparents" or
        "children".

        Ga and Gb can be used to mathematically calculate the relationship.

        .. seealso::
            http://en.wikipedia.org/wiki/Cousin#Mathematical_definitions

        :param Ga: The number of generations between the main person and the
                   common ancestor.
        :type Ga: int
        :param Gb: The number of generations between the group of people and the
                   common ancestor
        :type Gb: int
        :param reltocommon_a: relation path to common ancestor or common
                              Family for person a.
                              Note that length = Ga
        :type reltocommon_a: str
        :param reltocommon_b: relation path to common ancestor or common
                              Family for person b.
                              Note that length = Gb
        :type reltocommon_b: str
        :param only_birth: True if relation between a and b is by birth only
                           False otherwise
        :type only_birth: bool
        :param in_law_a: True if path to common ancestors is via the partner
                         of person a
        :type in_law_a: bool
        :param in_law_b: True if path to common ancestors is via the partner
                         of person b
        :type in_law_b: bool
        :returns: A string describing the relationship between the person and
                  the group.
        :rtype: str
        """
        rel_str = "verre familie"
        if Ga == 0:
            # These are descendants
            if Gb < len(_children_level):
                rel_str = _children_level[Gb]
            else:
                rel_str = "verre afstammelingen"
        elif Gb == 0:
            # These are parents/grand parents
            if Ga < len(_parents_level):
                rel_str = _parents_level[Ga]
            else:
                rel_str = "verre voorouders"
        elif Gb == 1:
            # These are siblings/aunts/uncles
            if Ga < len(_siblings_level):
                rel_str = _siblings_level[Ga]
            else:
                rel_str = "verre ooms/tantes"
        elif Ga == 1:
            # These are nieces/nephews
            if Gb < len(_nephews_nieces_level):
                rel_str = _nephews_nieces_level[Gb]
            else:
                rel_str = "verre neven/nichten"
        elif Ga > 1 and Ga == Gb:
            # These are cousins in the same generation
            if Ga <= len(_ordinal_level):
                rel_str = "%s neven" % _ordinal_level[Ga - 1]
            else:
                rel_str = "verre neven"
        elif Ga > 1 and Ga > Gb:
            # These are cousins in different generations with the second person
            # being in a higher generation from the common ancestor than the
            # first person.
            if Gb <= len(_LEVEL_NAME) and (Ga - Gb) < len(_removed_level):
                rel_str = "%s neven%s (omhoog)" % (
                    _ordinal_level[Gb - 1],
                    _removed_level[Ga - Gb],
                )
            else:
                rel_str = "verre neven"
        elif Gb > 1 and Gb > Ga:
            # These are cousins in different generations with the second person
            # being in a lower generation from the common ancestor than the
            # first person.
            if Ga <= len(_LEVEL_NAME) and (Gb - Ga) < len(_removed_level):
                rel_str = "%s neven%s (omlaag)" % (
                    _ordinal_level[Ga - 1],
                    _removed_level[Gb - Ga],
                )
            else:
                rel_str = "verre neven"

        if in_law_b is True:
            rel_str = "echtgenoten van %s" % rel_str

        return rel_str

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Return a string representing the relationship between the two people,
        see english method, eg b is father of a
        """
        if only_birth:
            step = ""
        else:
            step = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        rel_str = "verre %s%sfamilie" % (inlaw, step)

        if Gb == 0:
            # b is ancestor
            if Ga == 0:
                rel_str = "zelfde persoon"
            elif Ga == 1 and inlaw and not step:
                if gender_b == Person.MALE:
                    rel_str = "schoonvader"
                elif gender_b == Person.FEMALE:
                    rel_str = "schoonmoeder"
                else:
                    rel_str = "schoonouder"
            elif gender_b == Person.MALE:
                rel_str = self._get_father(Ga, step, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self._get_mother(Ga, step, inlaw)
            else:
                rel_str = self._get_parent_unknown(Ga, step, inlaw)
        elif Ga == 0:
            # a is descendant
            if Gb == 1 and inlaw and not step:
                if gender_b == Person.MALE:
                    rel_str = "schoonzoon"
                elif gender_b == Person.FEMALE:
                    rel_str = "schoondochter"
                else:
                    rel_str = "schoonkind"
            elif Gb == 1 and inlaw and step:
                # inlaw stepchild
                if gender_b == Person.MALE:
                    rel_str = "aangetrouwde stiefzoon"
                elif gender_b == Person.FEMALE:
                    rel_str = "aangetrouwde stiefdochter"
                else:
                    rel_str = "aangetrouwd stiefkind"
            elif gender_b == Person.MALE:
                rel_str = self._get_son(Gb, step, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self._get_daughter(Gb, step, inlaw)
            else:
                rel_str = self._get_child_unknown(Gb, step, inlaw)
        elif Ga > Gb:
            # b is higher in the branch, in english uncle/aunt or
            # cousin up, in Dutch always 'oom/tante'
            if gender_b == Person.MALE:
                rel_str = self._get_uncle(Ga - Gb, Gb, step, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self._get_aunt(Ga - Gb, Gb, step, inlaw)
            else:
                rel_str = (
                    self._get_uncle(Ga - Gb, Gb, step, inlaw)
                    + " of "
                    + self._get_aunt(Ga - Gb, Gb, step, inlaw)
                )
        elif Ga < Gb:
            # b is lower in the branch, in english niece/nephew or
            # cousin down, in Dutch always 'neef/nicht'
            if gender_b == Person.MALE:
                rel_str = self._get_nephew(Gb - Ga, Ga, step, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self._get_niece(Gb - Ga, Ga, step, inlaw)
            else:
                rel_str = (
                    self._get_nephew(Gb - Ga, Ga, step, inlaw)
                    + " of "
                    + self._get_niece(Gb - Ga, Ga, step, inlaw)
                )
        else:
            # people on the same level Ga == Gb
            if gender_b == Person.MALE:
                rel_str = self._get_male_cousin(Ga, step, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self._get_female_cousin(Ga, step, inlaw)
            else:
                rel_str = (
                    self._get_male_cousin(Ga, step, inlaw)
                    + " of "
                    + self._get_female_cousin(Ga, step, inlaw)
                )

        return rel_str

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        """
        Determine the string giving the relation between two siblings of
        type sib_type.
        Eg: b is the brother of a
        Here 'brother' is the string we need to determine
        This method gives more details about siblings than
        get_single_relationship_string can do.

        .. warning:: DON'T TRANSLATE THIS PROCEDURE IF LOGIC IS EQUAL IN YOUR
                     LANGUAGE, AND SAME METHODS EXIST (get_uncle, get_aunt,
                     get_sibling)
        """
        if sib_type == self.NORM_SIB or sib_type == self.UNKNOWN_SIB:
            typestr = ""
        elif sib_type == self.HALF_SIB_FATHER or sib_type == self.HALF_SIB_MOTHER:
            typestr = self.HALF
        elif sib_type == self.STEP_SIB:
            typestr = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if inlaw and not typestr:
            if gender_b == Person.MALE:
                rel_str = "schoonbroer"
            elif gender_b == Person.FEMALE:
                rel_str = "schoonzus"
            else:
                rel_str = "schoonbroer of -zus"
        else:
            if gender_b == Person.MALE:
                rel_str = self._get_male_cousin(1, typestr, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self._get_female_cousin(1, typestr, inlaw)
            else:
                rel_str = (
                    self._get_male_cousin(1, typestr, inlaw)
                    + " of "
                    + self._get_female_cousin(1, typestr, inlaw)
                )
        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    #    python src/plugins/rel/rel_nl.py

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#               2009       Benny Malengier
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python libraries
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

WEIGHT_NORMAL = Pango.Weight.NORMAL
WEIGHT_BOLD = Pango.Weight.BOLD

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from ...widgets.undoablebuffer import UndoableBuffer
from gramps.gen.display.name import displayer as name_displayer

# -------------------------------------------------------------------------
#
# NameModel
#
# -------------------------------------------------------------------------

YES = _("Yes")
NO = _("No")


class NameModel(Gtk.TreeStore):
    # tree groups defined
    DEFINDEX = 0
    DEFNAME = _("Preferred name")
    ALTINDEX = 1
    ALTNAME = _("Alternative names")

    _GROUPSTRING = _("%(groupname)s - %(groupnumber)d")

    COL_NAME = (0, str)
    COL_TYPE = (1, str)
    COL_DATA = (2, object)
    COL_FONTWEIGHT = (3, int)
    COL_GROUPAS = (4, str)
    COL_HASSOURCE = (5, bool)
    COL_NOTEPREVIEW = (6, str)
    COL_PRIVATE = (7, bool)

    COLS = (
        COL_NAME,
        COL_TYPE,
        COL_DATA,
        COL_FONTWEIGHT,
        COL_GROUPAS,
        COL_HASSOURCE,
        COL_NOTEPREVIEW,
        COL_PRIVATE,
    )

    def __init__(self, obj_list, db, groups):
        """
        @param obj_list: A list of lists, every entry is a group, the entries
            in a group are the data that needs to be shown subordinate to the
            group
        @param db: a database objects that can be used to obtain info
        @param groups: a list of (key, name) tuples. key is a key for the group
            that might be used. name is the name for the group.
        """
        typeobjs = (x[1] for x in self.COLS)
        Gtk.TreeStore.__init__(self, *typeobjs)
        self.db = db
        self.groups = groups
        for index, group in enumerate(obj_list):
            parentiter = self.append(None, row=self.row_group(index, group))
            for obj in group:
                self.append(parentiter, row=self.row(index, obj))

    def row_group(self, index, group):
        name = self.namegroup(index, len(group))
        return [name, "", (index, None), WEIGHT_NORMAL, "", "", "", None]

    def row(self, index, name):
        """
        Returns the row of the model in group index, and name as a
        list
        """
        return [
            name_displayer.display_name(name),
            str(name.get_type()),
            (index, name),
            self.colweight(index),
            name.get_group_as(),
            name.has_citations(),
            self.notepreview(name),
            name.get_privacy(),
        ]

    def colweight(self, index):
        if index == self.DEFINDEX:
            return WEIGHT_BOLD
        else:
            return WEIGHT_NORMAL

    def namegroup(self, groupindex, length):
        if groupindex == self.DEFINDEX:
            return self.DEFNAME
        return self._GROUPSTRING % {"groupname": self.ALTNAME, "groupnumber": length}

    def update_defname(self, defname):
        """
        callback if change to the preferred name happens
        """
        # default name is path (0,0)
        self.remove(self.get_iter((self.DEFINDEX, 0)))
        self.insert(
            self.get_iter(self.DEFINDEX), 0, row=self.row(self.DEFINDEX, defname)
        )

    def notepreview(self, name):
        nlist = name.get_note_list()
        if nlist:
            note = self.db.get_note_from_handle(nlist[0])
            text = note.get().replace("\n", " ")
            if len(text) > 80:
                text = text[:80] + "..."
            return text
        else:
            return ""



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import time
from .editors import EditFilter
from gramps.gen.const import CUSTOM_FILTERS
from gramps.gen.filters import (
    rules,
    FilterList,
    GenericFilterFactory,
    reload_custom_filters,
)
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from collections import abc


def make_filter(dbstate, uistate, objclass, gramps_ids, title=None):
    """
    Makes a Gramps Filter through dialog from a enumeration (list,
    set, etc.) of gramps_ids of type objclass.

    >>> make_filter(dbstate, uistate, 'Person', ['I0003', ...])
    """
    FilterClass = GenericFilterFactory(objclass)
    rule = getattr(getattr(rules, objclass.lower()), "RegExpIdOf")
    filter = FilterClass()
    if title is None:
        title = _("Filter %s from Clipboard") % objclass
    if isinstance(title, abc.Callable):
        title = title()
    filter.set_name(title)
    struct_time = time.localtime()
    filter.set_comment(
        _("Created on %(year)4d/%(month)02d/%(day)02d")
        % {
            "year": struct_time.tm_year,
            "month": struct_time.tm_mon,
            "day": struct_time.tm_mday,
        }
    )
    re = "|".join(["^%s$" % gid for gid in sorted(gramps_ids)])
    re_rule = rule([re])
    re_rule.use_regex = True
    filter.add_rule(re_rule)
    filterdb = FilterList(CUSTOM_FILTERS)
    filterdb.load()
    EditFilter(
        objclass,
        dbstate,
        uistate,
        [],
        filter,
        filterdb,
        lambda: edit_filter_save(uistate, filterdb, objclass),
    )


def edit_filter_save(uistate, filterdb, objclass):
    """
    If a filter changed, save them all. Reloads, and also calls callback.
    """
    filterdb.save()
    reload_custom_filters()
    uistate.emit("filters-changed", (objclass,))



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging
import cairo

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gramps.gen.config import config

# -------------------------------------------------------------------------
#
# osmGpsMap
#
# -------------------------------------------------------------------------

try:
    import gi
    from gi.repository import OsmGpsMap as osmgpsmap

    gi.require_version("OsmGpsMap", "1.0")
except:
    raise

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
_LOG = logging.getLogger("maps.datelayer")


class DateLayer(GObject.GObject, osmgpsmap.MapLayer):
    """
    This is the layer used to display the two extreme dates
    on the top left of the view
    """

    def __init__(self):
        """
        Initialize the layer
        """
        GObject.GObject.__init__(self)
        self.first = "    "
        self.last = "    "
        self.color = "black"
        self.font = config.get("utf8.selected-font")
        self.size = 36

    def clear_dates(self):
        """
        reset the layer attributes.
        """
        self.first = "    "
        self.last = "    "
        self.color = "black"
        self.font = config.get("utf8.selected-font")
        self.size = 36

    def set_font_attributes(self, font, size, color):
        """
        Set the font color, size and name
        """
        self.color = color
        if font:
            self.font = font
        else:
            self.font = config.get("utf8.selected-font")
        self.size = size

    def add_date(self, date):
        """
        Add a date
        """
        if date == "    " or date == "0000" or date == "9999":
            return
        if date < self.first or self.first == "    ":
            self.first = date
        if date > self.last or self.last == "    ":
            self.last = date

    def do_draw(self, gpsmap, ctx):
        """
        Draw the two extreme dates
        """
        dummy_map = gpsmap
        ctx.select_font_face(
            self.font, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL
        )
        ctx.set_font_size(int(self.size))
        color = Gdk.color_parse(self.color)
        ctx.set_source_rgba(
            float(color.red / 65535.0),
            float(color.green / 65535.0),
            float(color.blue / 65535.0),
            0.6,
        )  # transparency
        coord_x = 10
        coord_y = 15 + 2 * int(self.size)  # Display the oldest date
        ctx.move_to(coord_x, coord_y)
        ctx.show_text(self.first)
        coord_y = 15 + 3 * int(self.size)  # Display the newest date
        ctx.move_to(coord_x, coord_y)
        ctx.show_text(self.last)

    def do_render(self, gpsmap):
        """
        render the layer
        """
        pass

    def do_busy(self):
        """
        set the layer busy
        """
        return False

    def do_button_press(self, gpsmap, gdkeventbutton):
        """
        When we press a button.
        """
        dummy_map = gpsmap
        dummy_evt = gdkeventbutton
        return False


GObject.type_register(DateLayer)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2010       Raphael Ackermann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule


# -------------------------------------------------------------------------
#
# HasIdOf
#
# -------------------------------------------------------------------------
class HasGrampsId(Rule):
    """Rule that checks for an object with a specific Gramps ID."""

    labels = [_("ID:")]
    name = "Object with <Id>"
    description = "Matches objects with a specified Gramps ID"
    category = _("General filters")

    def apply(self, db, obj):
        """
        apply the rule on the obj.
        return true if the rule passes, false otherwise.
        """
        return obj.gramps_id == self.list[0]



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.utils.thumbnails import get_thumbnail_image, SIZE_NORMAL, SIZE_LARGE
from ..utils import is_right_click, open_file_with_default_application
from ..widgets.menuitem import add_menuitem
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Photo class
#
# -------------------------------------------------------------------------
class Photo(Gtk.EventBox):
    """
    Displays an image and allows it to be viewed in an external image viewer.
    """

    def __init__(self, use_small_size=False):
        Gtk.EventBox.__init__(self)
        self.full_path = None
        self.uistate = None
        self.handle = None
        self.photo = Gtk.Image()
        self.add(self.photo)
        self.connect("button-press-event", self.handle_button_press)
        tip = _(
            "Double-click on the picture to view it in the default image "
            "viewer application."
        )
        self.set_tooltip_text(tip)
        self.__size = SIZE_LARGE
        if use_small_size:
            self.__size = SIZE_NORMAL

    def set_image(self, full_path, mime_type=None, rectangle=None):
        """
        Set the image to be displayed.
        """
        self.full_path = full_path
        if full_path:
            pixbuf = get_thumbnail_image(full_path, mime_type, rectangle, self.__size)
            self.photo.set_from_pixbuf(pixbuf)
            self.photo.show()
        else:
            self.photo.hide()

    def handle_button_press(self, widget, event):
        """
        Display the image with the default external viewer.
        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            open_file_with_default_application(self.full_path, self.uistate)
            return True
        elif is_right_click(event):
            if self.handle and self.uistate:
                self.menu = Gtk.Menu()
                add_menuitem(
                    self.menu,
                    _("Make Active Media"),
                    widget,
                    lambda obj: self.uistate.set_active(self.handle, "Media"),
                )
                self.menu.popup_at_pointer(event)
                return True
        return False

    def set_uistate(self, uistate, handle):
        """
        Set uistate and media handle so that Photo can be handled by
        UI.
        """
        self.uistate = uistate
        self.handle = handle

    def set_pixbuf(self, full_path, pixbuf):
        """
        Set the image to be displayed from a pixbuf.
        """
        self.full_path = full_path
        if full_path:
            self.photo.set_from_pixbuf(pixbuf)
            self.photo.show()
        else:
            self.photo.hide()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from xml.sax import handler
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ._genericfilter import GenericFilterFactory
from . import rules


# -------------------------------------------------------------------------
#
# FilterParser
#
# -------------------------------------------------------------------------
class FilterParser(handler.ContentHandler):
    """Parses the XML file and builds the list of filters"""

    def __init__(self, gfilter_list):
        handler.ContentHandler.__init__(self)
        self.gfilter_list = gfilter_list
        self.f = None
        self.r = None
        self.a = []
        self.cname = None
        self.namespace = "Person"
        self.use_regex = False
        self.use_case = False

    def setDocumentLocator(self, locator):
        self.locator = locator

    def startElement(self, tag, attrs):
        if tag == "object":
            if "type" in attrs:
                self.namespace = attrs["type"]
            else:
                self.namespace = "generic"
            if self.namespace == "MediaObject":
                # deals with older custom filters
                self.namespace = "Media"
        elif tag == "filter":
            self.f = GenericFilterFactory(self.namespace)()
            self.f.set_name(attrs["name"])
            if "function" in attrs:
                try:
                    if int(attrs["function"]):
                        op = "or"
                    else:
                        op = "and"
                except ValueError:
                    op = attrs["function"]
                self.f.set_logical_op(op)
            if "invert" in attrs:
                self.f.set_invert(attrs["invert"])
            if "comment" in attrs:
                self.f.set_comment(attrs["comment"])
            self.gfilter_list.add(self.namespace, self.f)
        elif tag == "rule":
            if "use_regex" in attrs:
                self.use_regex = attrs["use_regex"] == "True"
            else:
                self.use_regex = False
            if "use_case" in attrs:
                self.use_case = attrs["use_case"] == "True"
            else:
                self.use_case = False
            save_name = attrs["class"]
            if save_name in old_names_2_class:
                self.r = old_names_2_class[save_name]
            else:
                try:
                    # First try to use fully qualified name
                    exec("self.r = %s" % save_name)
                except (ImportError, NameError, AttributeError):
                    # Now try to use name from rules.namespace
                    mc_match = save_name.split(".")
                    last_name = mc_match[-1]
                    try:
                        exec(
                            "self.r = rules.%s.%s" % (self.namespace.lower(), last_name)
                        )
                    except (ImportError, NameError, AttributeError):
                        print(
                            "ERROR: Filter rule '%s' in "
                            "filter '%s' not found!" % (save_name, self.f.get_name())
                        )
                        self.r = None
                        return
            self.a = []
        elif tag == "arg":
            self.a.append(attrs["value"])

    def endElement(self, tag):
        if tag == "rule" and self.r is not None:
            if len(self.r.labels) != len(self.a):
                self.__upgrade()
            if len(self.r.labels) < len(self.a):
                print(
                    _(
                        "WARNING: Too many arguments in filter '%s'!\n"
                        "Trying to load with subset of arguments."
                    )
                    % self.f.get_name()
                )
                nargs = len(self.r.labels)
                rule = self.r(self.a[0:nargs], self.use_regex, self.use_case)
                self.f.add_rule(rule)
            else:
                if len(self.r.labels) > len(self.a):
                    print(
                        _(
                            "WARNING: Too few arguments in filter '%s'!\n"
                            "         Trying to load anyway in the hope this "
                            "will be upgraded."
                        )
                        % self.f.get_name()
                    )
                try:
                    rule = self.r(self.a, self.use_regex, self.use_case)
                except AssertionError as msg:
                    print(msg)
                    print(
                        _(
                            "ERROR: filter %s could not be correctly loaded. "
                            "Edit the filter!"
                        )
                        % self.f.get_name()
                    )
                    return

                self.f.add_rule(rule)

    def characters(self, data):
        pass

    def __upgrade(self):
        """
        Upgrade argument lists to latest version.
        """
        # HasPlace rule has extra locality field in v3.3
        if self.r == rules.place.HasPlace and len(self.a) == 8:
            self.a = self.a[0:2] + [""] + self.a[4:8] + [self.a[3]] + [self.a[2]]
        # HasNameOf rule has new fields for surnames in v3.3
        if self.r == rules.person.HasNameOf and len(self.a) == 7:
            self.a = (
                self.a[0:2]
                + [self.a[3]]
                + [self.a[2]]
                + [self.a[6]]
                + [""]
                + [self.a[4]]
                + ["", ""]
                + [self.a[5]]
                + ["", "0"]
            )
        # New regular expression code in v3.4
        if (
            self.r
            in (
                rules.person.HasNameOf,
                rules.family.FatherHasNameOf,
                rules.family.MotherHasNameOf,
                rules.family.ChildHasNameOf,
            )
            and len(self.a) == 12
        ):
            self.use_regex = self.a[11] == "1"
            self.a = self.a[:11]
        if self.r == rules.person.HasTextMatchingSubstringOf and len(self.a) == 3:
            self.use_regex = self.a[2] == "1"
            self.a = self.a[:2]
        # HasEvent rule has extra primary role field in v3.4.7
        if self.r == rules.person.HasEvent and len(self.a) == 5:
            self.a.append("1")
        # IsEnclosedBy rule has extra inclusive field in v4.2.4
        if self.r == rules.place.IsEnclosedBy and len(self.a) == 1:
            self.a.append("0")


# -------------------------------------------------------------------------
#
# Name to class mappings
#
# -------------------------------------------------------------------------
# This dict is mapping from old names to new names, so that the existing
# custom_filters.xml will continue working
old_names_2_class = {
    "Everyone": rules.person.Everyone,
    "Is default person": rules.person.IsDefaultPerson,
    "Is bookmarked person": rules.person.IsBookmarked,
    "Has the Id": rules.person.HasIdOf,
    "Has a name": rules.person.HasNameOf,
    "Has the relationships": rules.person.HasRelationship,
    "Has the death": rules.person.HasDeath,
    "Has the birth": rules.person.HasBirth,
    "Is a descendant of": rules.person.IsDescendantOf,
    "Is a descendant family member of": rules.person.IsDescendantFamilyOf,
    "Is a descendant of filter match": rules.person.IsDescendantOfFilterMatch,
    "Is a descendant of person not more than N generations away": rules.person.IsLessThanNthGenerationDescendantOf,
    "Is a descendant of person at least N generations away": rules.person.IsMoreThanNthGenerationDescendantOf,
    "Is an descendant of person at least N generations away": rules.person.IsMoreThanNthGenerationDescendantOf,
    "Is a child of filter match": rules.person.IsChildOfFilterMatch,
    "Is an ancestor of": rules.person.IsAncestorOf,
    "Is an ancestor of filter match": rules.person.IsAncestorOfFilterMatch,
    "Is an ancestor of person not more than N generations away": rules.person.IsLessThanNthGenerationAncestorOf,
    "Is an ancestor of person at least N generations away": rules.person.IsMoreThanNthGenerationAncestorOf,
    "Is a parent of filter match": rules.person.IsParentOfFilterMatch,
    "Has a common ancestor with": rules.person.HasCommonAncestorWith,
    "Has a common ancestor with filter match": rules.person.HasCommonAncestorWithFilterMatch,
    "Is a female": rules.person.IsFemale,
    "Is a male": rules.person.IsMale,
    "Has the personal event": rules.person.HasEvent,
    "Has the family event": rules.person.HasFamilyEvent,
    "Has the personal attribute": rules.person.HasAttribute,
    "Has the family attribute": rules.person.HasFamilyAttribute,
    "Has source of": rules.person.HasSourceOf,
    "Matches the filter named": rules.person.HasSourceOf,
    "Is spouse of filter match": rules.person.IsSpouseOfFilterMatch,
    "Is a sibling of filter match": rules.person.IsSiblingOfFilterMatch,
    "Relationship path between two people": rules.person.RelationshipPathBetween,
    "Relationship paths between a person and a list of people": rules.person.DeepRelationshipPathBetween,
    "People who were adopted": rules.person.HaveAltFamilies,
    "People who have images": rules.person.HavePhotos,
    "People with children": rules.person.HaveChildren,
    "People with incomplete names": rules.person.IncompleteNames,
    "People with no marriage records": rules.person.NeverMarried,
    "People with multiple marriage records": rules.person.MultipleMarriages,
    "People without a birth date": rules.person.NoBirthdate,
    "People with incomplete events": rules.person.PersonWithIncompleteEvent,
    "Families with incomplete events": rules.person.FamilyWithIncompleteEvent,
    "People probably alive": rules.person.ProbablyAlive,
    "People marked private": rules.person.PeoplePrivate,
    "People marked public": rules.person.PeoplePublic,
    "Witnesses": rules.person.IsWitness,
    "Has text matching substring of": rules.person.HasTextMatchingSubstringOf,
}



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2015       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....lib import PlaceType


# -------------------------------------------------------------------------
#
# HasData
#
# -------------------------------------------------------------------------
class HasData(Rule):
    """
    Rule that checks for a place with a particular value
    """

    labels = [
        _("Name:", "place"),
        _("Place type:"),
        _("Code:"),
    ]
    name = _("Places matching parameters")
    description = _("Matches places with particular parameters")
    category = _("General filters")
    allow_regex = True

    def prepare(self, db, user):
        self.place_type = self.list[1]

        if self.place_type:
            self.place_type = PlaceType()
            self.place_type.set_from_xml_str(self.list[1])

    def apply(self, db, place):
        if not self.match_name(place):
            return False

        if self.place_type and place.get_type() != self.place_type:
            return False

        if not self.match_substring(2, place.get_code()):
            return False

        return True

    def match_name(self, place):
        """
        Match any name in a list of names.
        """
        for name in place.get_all_names():
            if self.match_substring(0, name.get_value()):
                return True
        return False



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re
import math
import logging
import gi
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.errors import WindowActiveError
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.dialog import WarningDialog
from gramps.gen.utils.location import get_main_location
from gramps.gen.lib import PlaceType
from gramps.gen.utils.place import conv_lat_lon
from gramps.gen.display.place import displayer as _pd
from .osmgps import OsmGps

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
try:
    gi.require_version("GeocodeGlib", "1.0")
    from gi.repository import GeocodeGlib

    GEOCODEGLIB = True
except:
    GEOCODEGLIB = False

# -------------------------------------------------------------------------
#
# Functions and variables
#
# -------------------------------------------------------------------------
PLACE_REGEXP = re.compile('<span background="green">(.*)</span>')
PLACE_STRING = '<span background="green">%s</span>'
GEOCODE_REGEXP = re.compile('<span background="red">(.*)</span>')
GEOCODE_STRING = '<span background="red">%s</span>'

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
_LOG = logging.getLogger("maps.placeselection")
_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# PlaceSelection
#
# -------------------------------------------------------------------------
class PlaceSelection(ManagedWindow, OsmGps):
    """
    We show a selection box for possible places in a region of the map.
    We can select the diameter of the region which is a circle.
    Depending of this region, we can show the possible choice.
    We select the value depending of our need which open the EditPlace box.
    """

    def __init__(
        self, uistate, dbstate, maps, layer, places, lat, lon, function, oldvalue=None
    ):
        """
        Place Selection initialization
        """
        try:
            ManagedWindow.__init__(self, uistate, [], PlaceSelection)
        except WindowActiveError:
            return
        OsmGps.__init__(self, uistate)
        self.uistate = uistate
        self.dbstate = dbstate
        self.places = []
        self.lat = lat
        self.lon = lon
        self.osm = maps
        self.radius = 1.0
        self.circle = None
        self.oldvalue = oldvalue
        self.place_list = places
        self.function = function
        self.selection_layer = layer
        self.layer = layer
        self.warning = False
        dlg = Gtk.Dialog(
            title=_("Place Selection in a region"), transient_for=uistate.window
        )
        dlg.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_window(dlg, None, _("Place Selection in a region"), None)
        mylabel = _(
            "Choose the radius of the selection.\n"
            "On the map you should see a circle or an"
            " oval depending on the latitude."
        )
        mylabel += _(
            "\nIn the following table you may have :"
            "\n - a green row related to a selected place."
        )
        if GEOCODEGLIB:
            mylabel += _("\n - a red row related to a geocoding result.")
        label = Gtk.Label(label=mylabel)
        label.set_valign(Gtk.Align.END)
        self.window.vbox.pack_start(label, False, True, 0)
        adj = Gtk.Adjustment(
            value=1.0,
            lower=0.1,
            upper=3.0,
            step_increment=0.1,
            page_increment=0,
            page_size=0,
        )
        # default value is 1.0, minimum is 0.1 and max is 3.0
        slider = Gtk.Scale(orientation=Gtk.Orientation.HORIZONTAL, adjustment=adj)
        slider.set_digits(1)
        slider.set_value_pos(Gtk.PositionType.BOTTOM)
        slider.connect("value-changed", self.slider_change, self.lat, self.lon)
        self.window.vbox.pack_start(slider, False, True, 0)
        self.vadjust = Gtk.Adjustment(page_size=15)
        self.scroll = Gtk.ScrolledWindow(vadjustment=self.vadjust)
        self.scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.scroll.set_shadow_type(Gtk.ShadowType.IN)
        self.plist = Gtk.ListStore(str, str, str, str, str)
        self.choices = Gtk.TreeView(model=self.plist)
        self.scroll.add(self.choices)
        self.renderer = Gtk.CellRendererText()
        self.tvcol1 = Gtk.TreeViewColumn(_("Country"), self.renderer, markup=0)
        self.tvcol2 = Gtk.TreeViewColumn(_("State"), self.renderer, markup=1)
        self.tvcol3 = Gtk.TreeViewColumn(_("County"), self.renderer, markup=2)
        self.tvcol4 = Gtk.TreeViewColumn(_("Other"), self.renderer, markup=3)
        self.tvcol1.set_sort_column_id(0)
        self.tvcol2.set_sort_column_id(1)
        self.tvcol3.set_sort_column_id(2)
        self.tvcol4.set_sort_column_id(3)
        self.choices.append_column(self.tvcol1)
        self.choices.append_column(self.tvcol2)
        self.choices.append_column(self.tvcol3)
        self.choices.append_column(self.tvcol4)
        self.window.vbox.pack_start(self.scroll, True, True, 0)
        self.label2 = Gtk.Label()
        self.label2.set_markup(
            '<span background="green" foreground="black"'
            ">%s</span>"
            % _(
                "The green values in the row correspond " "to the current place values."
            )
        )
        self.label2.set_valign(Gtk.Align.END)
        self.window.vbox.pack_start(self.label2, False, True, 0)
        self.window.set_default_size(400, 300)
        self.choices.connect("row-activated", self.selection, function)
        self.window.connect("response", self.close)
        self.window.show_all()
        self.show()
        self.label2.hide()
        self.slider_change(None, lat, lon)

    def close(self, *obj):
        """
        Close the selection place editor
        """
        self.hide_the_region()
        ManagedWindow.close(self, *obj)

    def slider_change(self, obj, lat, lon):
        """
        Display on the map a circle in which we select all the places
        inside this region.
        """
        self.radius = obj.get_value() if obj else 1.0
        self.show_the_region(self.radius)
        self.match(lat, lon)
        self.plist.clear()
        if self.oldvalue != None:
            # The old values are always in the first row.
            # In this case, we change the color of the row.
            # display the associated message
            self.label2.show()
            place = self.dbstate.db.get_place_from_handle(self.oldvalue)
            loc = get_main_location(self.dbstate.db, place)
            self.plist.append(
                (
                    PLACE_STRING % loc.get(PlaceType.COUNTRY, ""),
                    PLACE_STRING % loc.get(PlaceType.STATE, ""),
                    PLACE_STRING % loc.get(PlaceType.COUNTY, ""),
                    PLACE_STRING % _("Other"),
                    self.oldvalue,
                )
            )
        for place in self.places:
            self.plist.append((place[0], place[1], place[2], place[3], place[4]))
        # here, we could add value from geography names services ...
        if GEOCODEGLIB:
            loc = GeocodeGlib.Location.new(lat, lon, 0)
            obj = GeocodeGlib.Reverse.new_for_location(loc)
            try:
                result = GeocodeGlib.Reverse.resolve(obj)
                self.plist.append(
                    (
                        GEOCODE_STRING % result.get_country(),
                        GEOCODE_STRING % result.get_state(),
                        GEOCODE_STRING % result.get_town(),
                        GEOCODE_STRING % result.get_name(),
                        "",
                    )
                )
            except:
                pass

        # if we found no place, we must create a default place.
        self.plist.append((_("New place with empty fields"), "", "...", "", None))

    def hide_the_region(self):
        """
        Hide the layer which contains the circle
        """
        layer = self.get_selection_layer()
        if layer:
            self.remove_layer(layer)

    def show_the_region(self, rds):
        """
        Show a circle in which we select the places.
        """
        # circle (rds)
        self.hide_the_region()
        self.selection_layer = self.add_selection_layer()
        self.selection_layer.add_circle(rds / 2.0, self.lat, self.lon)

    def get_location(self, gramps_id):
        """
        get location values
        """
        parent_place = None
        country = state = county = other = ""
        place = self.dbstate.db.get_place_from_gramps_id(gramps_id)
        place_name = place.name.get_value()
        parent_list = place.get_placeref_list()
        while parent_list:
            place = self.dbstate.db.get_place_from_handle(parent_list[0].ref)
            parent_list = place.get_placeref_list()
            if int(place.get_type()) == PlaceType.COUNTY:
                county = place.name.get_value()
                if parent_place is None:
                    parent_place = place.get_handle()
            elif int(place.get_type()) == PlaceType.STATE:
                state = place.name.get_value()
                if parent_place is None:
                    parent_place = place.get_handle()
            elif int(place.get_type()) == PlaceType.COUNTRY:
                country = place.name.get_value()
                if parent_place is None:
                    parent_place = place.get_handle()
            else:
                other = place.name.get_value()
                if parent_place is None:
                    parent_place = place.get_handle()
        return (country, state, county, place_name, other)

    def match(self, lat, lon):
        """
        coordinates matching.
        """
        rds = float(self.radius)
        self.places = []

        # place
        for entry in self.place_list:
            if math.hypot(lat - float(entry[3]), lon - float(entry[4])) <= rds:
                # Do we already have this place ? avoid duplicates
                (country, state, county, place, other) = self.get_location(entry[9])
                if not [country, state, county, place, other] in self.places:
                    self.places.append([country, state, county, place, other])
        self.warning = False
        for place in self.dbstate.db.iter_places():
            latn = place.get_latitude()
            lonn = place.get_longitude()
            if latn and lonn:
                latn, dummy_ignore = conv_lat_lon(latn, "0", "D.D8")
                if not latn:
                    if not self.warning:
                        self.close()
                    warn1 = _("you have a wrong latitude for:")
                    warn2 = _pd.display(self.dbstate.db, place) + "\n\n<b>"
                    warn2 += _("Please, correct this before linking") + "</b>"
                    WarningDialog(warn1, warn2, parent=self.uistate.window)
                    self.warning = True
                    continue
                dummy_ignore, lonn = conv_lat_lon("0", lonn, "D.D8")
                if not lonn:
                    if not self.warning:
                        self.close()
                    warn1 = _("you have a wrong longitude for:") + "\n"
                    warn2 = _pd.display(self.dbstate.db, place) + "\n\n<b>"
                    warn2 += _("Please, correct this before linking") + "</b>"
                    WarningDialog(warn1, warn2, parent=self.uistate.window)
                    self.warning = True
                    continue
                if math.hypot(lat - float(latn), lon - float(lonn)) <= rds:
                    (country, state, county, place, other) = self.get_location(
                        place.get_gramps_id()
                    )
                    if not [country, state, county, place, other] in self.places:
                        self.places.append([country, state, county, place, other])

    def selection(self, obj, index, column, function):
        """
        get location values and call the real function : add_place, edit_place
        """
        dummy_obj = obj
        dummy_col = column
        dummy_fct = function
        # TODO : self.plist unsubscriptable
        self.function(self.plist[index], self.lat, self.lon)

    def untag_text(self, text, tag):
        """
        suppress the green or red color tag.
        if tag = 0 : PLACE_REGEXP
        if tag = 1 : GEOCODE_REGEXP
        """

        if tag:
            regtag = GEOCODE_REGEXP
        else:
            regtag = PLACE_REGEXP

        match = regtag.match(text)
        if match:
            without_tags = match.groups()[0]
            return without_tags
        return text



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from math import pi
import logging

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# osmGpsMap
#
# -------------------------------------------------------------------------
from gi.repository import GObject

try:
    import gi

    gi.require_version("OsmGpsMap", "1.0")
    from gi.repository import OsmGpsMap as osmgpsmap
except:
    raise

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
_LOG = logging.getLogger("maps.selectionlayer")


class SelectionLayer(GObject.GObject, osmgpsmap.MapLayer):
    """
    This class is used to select an area on the map
    """

    def __init__(self):
        """
        Initialize thz selection layer
        """
        GObject.GObject.__init__(self)
        self.circles = []
        self.rectangles = []

    def add_circle(self, rds, lat, lon):
        """
        Add a circle
        """
        self.circles.append((rds, lat, lon))

    def add_rectangle(self, cp1, cp2):
        """
        Add a rectangle
        """
        self.rectangles.append((cp1, cp2))

    def do_draw(self, gpsmap, ctx):
        """
        draw the circles and the rectangles
        """
        for circle in self.circles:
            top_left = osmgpsmap.MapPoint.new_degrees(
                circle[1] + circle[0], circle[2] - circle[0]
            )
            bottom_right = osmgpsmap.MapPoint.new_degrees(
                circle[1] - circle[0], circle[2] + circle[0]
            )
            center = osmgpsmap.MapPoint.new_degrees(circle[1], circle[2])
            crd_x, crd_y = gpsmap.convert_geographic_to_screen(top_left)
            crd_x2, crd_y2 = gpsmap.convert_geographic_to_screen(bottom_right)
            crd_xc, crd_yc = gpsmap.convert_geographic_to_screen(center)
            width = float(crd_x2 - crd_x)
            height = float(crd_y2 - crd_y)
            ctx.save()
            ctx.translate(crd_xc, crd_yc)
            ctx.set_line_width(3.0)
            ctx.set_source_rgba(0.0, 0.0, 0.0, 0.8)
            ctx.scale(1.0, (height / width))
            ctx.arc(float(0.0), float(0.0), width, 0.0, 2 * pi)
            ctx.stroke()
            ctx.restore()

        for rectangle in self.rectangles:
            top_left, bottom_right = rectangle
            ctx.set_source_rgba(0.0, 0.0, 0.0, 0.8)
            ctx.set_line_width(3.0)
            crd_x, crd_y = gpsmap.convert_geographic_to_screen(top_left)
            crd_x2, crd_y2 = gpsmap.convert_geographic_to_screen(bottom_right)
            # be sure when can select a region in all case.
            if crd_x < crd_x2:
                if crd_y < crd_y2:
                    ctx.rectangle(crd_x, crd_y, crd_x2 - crd_x, crd_y2 - crd_y)
                else:
                    ctx.rectangle(crd_x, crd_y2, crd_x2 - crd_x, crd_y - crd_y2)
            else:
                if crd_y < crd_y2:
                    ctx.rectangle(crd_x2, crd_y, crd_x - crd_x2, crd_y2 - crd_y)
                else:
                    ctx.rectangle(crd_x2, crd_y2, crd_x - crd_x2, crd_y - crd_y2)
            ctx.stroke()

    def do_render(self, gpsmap):
        """
        render the layer
        """
        pass

    def do_busy(self):
        """
        set the map busy
        """
        return False

    def do_button_press(self, gpsmap, gdkeventbutton):
        """
        Someone press a button
        """
        dummy_map = gpsmap
        dummy_evt = gdkeventbutton
        return False


GObject.type_register(SelectionLayer)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2002       Gary Shao
# Copyright (C) 2007       Brian G. Matherly
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2014       Nick Hall
# Copyright (C) 2017       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import os
from xml.sax.saxutils import escape


def escxml(string):
    """
    Escapes XML special characters.
    """
    return escape(string, {'"': "&quot;"})


# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .paragraphstyle import ParagraphStyle
from .fontstyle import FontStyle
from .tablestyle import TableStyle, TableCellStyle
from .graphicstyle import GraphicsStyle

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".stylesheet")

# -------------------------------------------------------------------------
#
# SAX interface
#
# -------------------------------------------------------------------------
from xml.sax import make_parser, handler, SAXParseException


# ------------------------------------------------------------------------
#
# cnv2color
#
# ------------------------------------------------------------------------
def cnv2color(text):
    """
    converts a hex value in the form of #XXXXXX into a tuple of integers
    representing the RGB values
    """
    return (int(text[1:3], 16), int(text[3:5], 16), int(text[5:7], 16))


# ------------------------------------------------------------------------
#
# StyleSheetList
#
# ------------------------------------------------------------------------
class StyleSheetList:
    """
    Interface into the user's defined style sheets. Each StyleSheetList
    has a predefined default style specified by the report. Additional
    styles are loaded from a specified XML file if it exists.
    """

    def __init__(self, filename, defstyle):
        """
        Create a new StyleSheetList from the specified default style and
        any other styles that may be defined in the specified file.

        file - XML file that contains style definitions
        defstyle - default style
        """
        defstyle.set_name("default")
        self.map = {"default": defstyle}
        self.__file = filename
        self.parse()

    def delete_style_sheet(self, name):
        """
        Remove a style from the list. Since each style must have a
        unique name, the name is used to delete the stylesheet.

        name - Name of the style to delete
        """
        del self.map[name]

    def get_style_sheet_map(self):
        """
        Return the map of names to styles.
        """
        return self.map

    def get_style_sheet(self, name):
        """
        Return the StyleSheet associated with the name

        name - name associated with the desired StyleSheet.
        """
        return self.map[name]

    def get_style_names(self):
        "Return a list of all the style names in the StyleSheetList"
        return list(self.map.keys())

    def set_style_sheet(self, name, style):
        """
        Add or replaces a StyleSheet in the StyleSheetList. The
        default style may not be replaced.

        name - name associated with the StyleSheet to add or replace.
        style - definition of the StyleSheet
        """
        style.set_name(name)
        if name != "default":
            self.map[name] = style

    def save(self):
        """
        Saves the current StyleSheet definitions to the associated file.
        """
        with open(self.__file, "w") as xml_file:
            xml_file.write('<?xml version="1.0" encoding="utf-8"?>\n')
            xml_file.write("<stylelist>\n")

            for name in sorted(self.map.keys()):  # enable diff of archived ones
                if name == "default":
                    continue
                sheet = self.map[name]
                xml_file.write('  <sheet name="%s">\n' % escxml(name))

                for p_name in sorted(sheet.get_paragraph_style_names()):
                    self.write_paragraph_style(xml_file, sheet, p_name)

                for t_name in sorted(sheet.get_table_style_names()):
                    self.write_table_style(xml_file, sheet, t_name)

                for c_name in sorted(sheet.get_cell_style_names()):
                    self.write_cell_style(xml_file, sheet, c_name)

                for g_name in sorted(sheet.get_draw_style_names()):
                    self.write_graphics_style(xml_file, sheet, g_name)

                xml_file.write("  </sheet>\n")
            xml_file.write("</stylelist>\n")

    def write_paragraph_style(self, xml_file, sheet, p_name):
        para = sheet.get_paragraph_style(p_name)

        # Get variables for substitutions
        font = para.get_font()
        rmargin = float(para.get_right_margin())
        lmargin = float(para.get_left_margin())
        findent = float(para.get_first_indent())
        tmargin = float(para.get_top_margin())
        bmargin = float(para.get_bottom_margin())
        padding = float(para.get_padding())
        bg_color = para.get_background_color()

        # Write out style definition
        xml_file.write(
            '    <style name="%s">\n' % escxml(p_name)
            + "      <font "
            + 'face="%d" ' % font.get_type_face()
            + 'size="%d" ' % font.get_size()
            + 'italic="%d" ' % font.get_italic()
            + 'bold="%d" ' % font.get_bold()
            + 'underline="%d" ' % font.get_underline()
            + 'color="#%02x%02x%02x" ' % font.get_color()
            + "/>\n"
            + "      <para "
            + 'description="%s" ' % escxml(para.get_description())
            + 'rmargin="%.3f" ' % rmargin
            + 'lmargin="%.3f" ' % lmargin
            + 'first="%.3f" ' % findent
            + 'tmargin="%.3f" ' % tmargin
            + 'bmargin="%.3f" ' % bmargin
            + 'pad="%.3f" ' % padding
            + 'bgcolor="#%02x%02x%02x" ' % bg_color
            + 'level="%d" ' % para.get_header_level()
            + 'align="%d" ' % para.get_alignment()
            + 'tborder="%d" ' % para.get_top_border()
            + 'lborder="%d" ' % para.get_left_border()
            + 'rborder="%d" ' % para.get_right_border()
            + 'bborder="%d" ' % para.get_bottom_border()
            + "/>\n"
            + "    </style>\n"
        )

    def write_table_style(self, xml_file, sheet, t_name):
        t_style = sheet.get_table_style(t_name)

        # Write out style definition
        xml_file.write(
            '    <style name="%s">\n' % escxml(t_name)
            + "      <table "
            + 'description="%s" ' % escxml(t_style.get_description())
            + 'width="%d" ' % t_style.get_width()
            + 'columns="%d"' % t_style.get_columns()
            + ">\n"
        )

        for col in range(t_style.get_columns()):
            column_width = t_style.get_column_width(col)
            xml_file.write('        <column width="%d" />\n' % column_width)

        xml_file.write("      </table>\n")
        xml_file.write("    </style>\n")

    def write_cell_style(self, xml_file, sheet, c_name):
        cell = sheet.get_cell_style(c_name)

        # Write out style definition
        xml_file.write(
            '    <style name="%s">\n' % escxml(c_name)
            + "      <cell "
            + 'description="%s" ' % escxml(cell.get_description())
            + 'lborder="%d" ' % cell.get_left_border()
            + 'rborder="%d" ' % cell.get_right_border()
            + 'tborder="%d" ' % cell.get_top_border()
            + 'bborder="%d" ' % cell.get_bottom_border()
            + 'pad="%.3f" ' % cell.get_padding()
            + "/>\n"
            + "    </style>\n"
        )

    def write_graphics_style(self, xml_file, sheet, g_name):
        draw = sheet.get_draw_style(g_name)

        # Write out style definition
        xml_file.write(
            '    <style name="%s">\n' % escxml(g_name)
            + "      <draw "
            + 'description="%s" ' % escxml(draw.get_description())
            + 'para="%s" ' % draw.get_paragraph_style()
            + 'width="%.3f" ' % draw.get_line_width()
            + 'style="%d" ' % draw.get_line_style()
            + 'color="#%02x%02x%02x" ' % draw.get_color()
            + 'fillcolor="#%02x%02x%02x" ' % draw.get_fill_color()
            + 'shadow="%d" ' % draw.get_shadow()
            + 'space="%.3f" ' % draw.get_shadow_space()
            + "/>\n"
            + "    </style>\n"
        )

    def parse(self):
        """
        Loads the StyleSheets from the associated file, if it exists.
        """
        try:
            if os.path.isfile(self.__file):
                parser = make_parser()
                parser.setContentHandler(SheetParser(self))
                with open(self.__file) as the_file:
                    parser.parse(the_file)
        except (IOError, OSError, SAXParseException):
            pass


# ------------------------------------------------------------------------
#
# StyleSheet
#
# ------------------------------------------------------------------------
class StyleSheet:
    """
    A collection of named paragraph styles.
    """

    def __init__(self, obj=None):
        """
        Create a new empty StyleSheet.

        :param obj: if not None, creates the StyleSheet from the values in
                    obj, instead of creating an empty StyleSheet
        """
        self.para_styles = {}
        self.draw_styles = {}
        self.table_styles = {}
        self.cell_styles = {}
        self.name = ""
        if obj is not None:
            for style_name, style in obj.para_styles.items():
                self.para_styles[style_name] = ParagraphStyle(style)
            for style_name, style in obj.draw_styles.items():
                self.draw_styles[style_name] = GraphicsStyle(style)
            for style_name, style in obj.table_styles.items():
                self.table_styles[style_name] = TableStyle(style)
            for style_name, style in obj.cell_styles.items():
                self.cell_styles[style_name] = TableCellStyle(style)

    def set_name(self, name):
        """
        Set the name of the StyleSheet

        :param name: The name to be given to the StyleSheet
        """
        self.name = name

    def get_name(self):
        """
        Return the name of the StyleSheet
        """
        return self.name

    def clear(self):
        "Remove all styles from the StyleSheet"
        self.para_styles = {}
        self.draw_styles = {}
        self.table_styles = {}
        self.cell_styles = {}

    def is_empty(self):
        "Checks if any styles are defined"
        style_count = (
            len(self.para_styles)
            + len(self.draw_styles)
            + len(self.table_styles)
            + len(self.cell_styles)
        )
        if style_count > 0:
            return False
        else:
            return True

    def add_paragraph_style(self, name, style):
        """
        Add a paragraph style to the style sheet.

        :param name: The name of the :class:`.ParagraphStyle`
        :param style: :class:`.ParagraphStyle` instance to be added.
        """
        self.para_styles[name] = ParagraphStyle(style)

    def get_paragraph_style(self, name):
        """
        Return the :class:`.ParagraphStyle` associated with the name

        :param name: name of the :class:`.ParagraphStyle` that is wanted
        """
        return ParagraphStyle(self.para_styles[name])

    def get_paragraph_style_names(self):
        "Return the list of paragraph names in the StyleSheet"
        return list(self.para_styles.keys())

    def add_draw_style(self, name, style):
        """
        Add a draw style to the style sheet.

        :param name: The name of the :class:`.GraphicsStyle`
        :param style: :class:`.GraphicsStyle` instance to be added.
        """
        self.draw_styles[name] = GraphicsStyle(style)

    def get_draw_style(self, name):
        """
        Return the :class:`.GraphicsStyle` associated with the name

        :param name: name of the :class:`.GraphicsStyle` that is wanted
        """
        return GraphicsStyle(self.draw_styles[name])

    def get_draw_style_names(self):
        "Return the list of draw style names in the StyleSheet"
        return list(self.draw_styles.keys())

    def add_table_style(self, name, style):
        """
        Add a table style to the style sheet.

        :param name: The name of the :class:`.TableStyle`
        :param style: :class:`.TableStyle` instance to be added.
        """
        self.table_styles[name] = TableStyle(style)

    def get_table_style(self, name):
        """
        Return the :class:`.TableStyle` associated with the name

        :param name: name of the :class:`.TableStyle` that is wanted
        """
        return TableStyle(self.table_styles[name])

    def get_table_style_names(self):
        "Return the list of table style names in the StyleSheet"
        return list(self.table_styles.keys())

    def add_cell_style(self, name, style):
        """
        Add a cell style to the style sheet.

        :param name: The name of the :class:`.TableCellStyle`
        :param style: :class:`.TableCellStyle` instance to be added.
        """
        self.cell_styles[name] = TableCellStyle(style)

    def get_cell_style(self, name):
        """
        Return the :class:`.TableCellStyle` associated with the name

        :param name: name of the :class:`.TableCellStyle` that is wanted
        """
        return TableCellStyle(self.cell_styles[name])

    def get_cell_style_names(self):
        "Return the list of cell style names in the StyleSheet"
        return list(self.cell_styles.keys())


# -------------------------------------------------------------------------
#
# SheetParser
#
# -------------------------------------------------------------------------
class SheetParser(handler.ContentHandler):
    """
    SAX parsing class for the StyleSheetList XML file.
    """

    def __init__(self, sheetlist):
        """
        Create a SheetParser class that populates the passed StyleSheetList
        class.

        sheetlist - :class:`StyleSheetList` instance to be loaded from the file.
        """
        handler.ContentHandler.__init__(self)
        self.sheetlist = sheetlist
        self.f = None
        self.p = None
        self.s = None
        self.t = None
        self.columns_widths = []
        self.sheet_name = None
        self.style_name = None

    def startElement(self, tag, attrs):
        """
        Overridden class that handles the start of a XML element
        """
        if tag == "sheet":
            self.s = StyleSheet(self.sheetlist.map["default"])
            self.sheet_name = attrs["name"]
        elif tag == "font":
            self.f = FontStyle()
            self.f.set_type_face(int(attrs["face"]))
            self.f.set_size(int(attrs["size"]))
            self.f.set_italic(int(attrs["italic"]))
            self.f.set_bold(int(attrs["bold"]))
            self.f.set_underline(int(attrs["underline"]))
            self.f.set_color(cnv2color(attrs["color"]))
        elif tag == "para":
            self.p = ParagraphStyle()
            if "description" in attrs:
                self.p.set_description(attrs["description"])
            self.p.set_right_margin(float(attrs["rmargin"]))
            self.p.set_left_margin(float(attrs["lmargin"]))
            self.p.set_first_indent(float(attrs["first"]))
            try:
                # This is needed to read older style files
                # lacking tmargin and bmargin
                self.p.set_top_margin(float(attrs["tmargin"]))
                self.p.set_bottom_margin(float(attrs["bmargin"]))
            except KeyError:
                pass
            self.p.set_padding(float(attrs["pad"]))
            self.p.set_alignment(int(attrs["align"]))
            self.p.set_right_border(int(attrs["rborder"]))
            self.p.set_header_level(int(attrs["level"]))
            self.p.set_left_border(int(attrs["lborder"]))
            self.p.set_top_border(int(attrs["tborder"]))
            self.p.set_bottom_border(int(attrs["bborder"]))
            self.p.set_background_color(cnv2color(attrs["bgcolor"]))
            self.p.set_font(self.f)
        elif tag == "style":
            self.style_name = attrs["name"]
        elif tag == "table":
            self.t = TableStyle()
            if "description" in attrs:
                self.t.set_description(attrs["description"])
            self.t.set_width(int(attrs["width"]))
            self.t.set_columns(int(attrs["columns"]))
            self.column_widths = []
        elif tag == "column":
            self.column_widths.append(int(attrs["width"]))
        elif tag == "cell":
            self.c = TableCellStyle()
            if "description" in attrs:
                self.c.set_description(attrs["description"])
            self.c.set_left_border(int(attrs["lborder"]))
            self.c.set_right_border(int(attrs["rborder"]))
            self.c.set_top_border(int(attrs["tborder"]))
            self.c.set_bottom_border(int(attrs["bborder"]))
            self.c.set_padding(float(attrs["pad"]))
        elif tag == "draw":
            self.g = GraphicsStyle()
            if "description" in attrs:
                self.g.set_description(attrs["description"])
            self.g.set_paragraph_style(attrs["para"])
            self.g.set_line_width(float(attrs["width"]))
            self.g.set_line_style(int(attrs["style"]))
            self.g.set_color(cnv2color(attrs["color"]))
            self.g.set_fill_color(cnv2color(attrs["fillcolor"]))
            self.g.set_shadow(int(attrs["shadow"]), float(attrs["space"]))

    def endElement(self, tag):
        "Overridden class that handles the end of a XML element"
        if tag == "sheet":
            self.sheetlist.set_style_sheet(self.sheet_name, self.s)
        elif tag == "para":
            self.s.add_paragraph_style(self.style_name, self.p)
        elif tag == "table":
            self.t.set_column_widths(self.column_widths)
            self.s.add_table_style(self.style_name, self.t)
        elif tag == "cell":
            self.s.add_cell_style(self.style_name, self.c)
        elif tag == "draw":
            self.s.add_draw_style(self.style_name, self.g)



# encoding:utf-8
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010 Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# plugins/webstuff/webstuff.gpr.py

MODULE_VERSION = "5.2"

# ------------------------------------------------------------------------
#
# Stylesheets
#
# ------------------------------------------------------------------------
register(
    GENERAL,
    id="system webstuff",
    category="WEBSTUFF",
    name=_("Webstuff"),
    description=_("Provides a collection of resources for the web"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    fname="webstuff.py",
    load_on_reg=True,
    process="process_list",
    status=STABLE,
)
